EXT_TO_TREE_SITTER_LANGUAGE = {
    ".js": "tsx",
    ".jsx": "tsx",
    ".ts": "tsx",
    ".tsx": "tsx",
    ".mjs": "tsx",
    ".py": "python",
    ".rs": "rust",
    ".go": "go",
    ".java": "java",
    ".cpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    ".c": "cpp",
    ".h": "cpp",
    ".hpp": "cpp",
    ".cs": "cpp",
    ".rb": "ruby",
    # ".md": "markdown",
    # ".rst": "markdown",
    # ".txt": "markdown",
    # ".erb": "embedded-template",
    # ".ejs": "embedded-template",
    # ".html": "embedded-template",
    ".erb": "html",
    ".ejs": "html",
    ".html": "html",
    ".vue": "html",
    ".php": "php",
}

NAME_TO_TREE_SITTER_LANGUAGE = {
    "javascript": "tsx",
    "typescript": "tsx",
    "python": "python",
    "rust": "rust",
    "go": "go",
    "java": "java",
    "c": "cpp",
    "cpp": "cpp",
    "ruby": "ruby",
    "csharp": "cpp",
    "html": "html",
    "php": "php",
}



# Standard library
import os
from enum import Enum
from dataclasses import dataclass
from typing import Tuple, List
import shutil
import tempfile

# Local
from withrepo.resources.languages import EXT_TO_LANGUAGE_DATA
from withrepo.constants import LANGUAGE_TO_LSP_LANGUAGE_MAP


# General utils
def flatten(xss):
    return [x for xs in xss for x in xs]


def is_file_empty(path):
    return os.stat(path).st_size == 0


# Dtos for downloads
class RepoProvider(Enum):
    GITHUB = "github"
    GITLAB = "gitlab"
    BITBUCKET = "bitbucket"


@dataclass
class RepoArguments:
    user: str = ""
    repo: str = ""
    commit: str = ""
    url: str = ""
    branch: str = ""
    provider: RepoProvider = None
    root_dir: str = ""
    root_path: str = None

    def invalid(self) -> bool:
        return not any(
            [self.user, self.repo, self.commit, self.url, self.branch, self.root_path]
        )


@dataclass
class LanguageGroup:
    language: str
    path: str
    # files: List[str]


# Codebase segmentation utils
def get_all_paths_from_root_relative(root_path):
    abs_paths, rel_paths = [], []
    for root, dirs, files in os.walk(root_path):
        for file in files:
            abs_path = os.path.join(root, file)
            relpath = os.path.relpath(abs_path, root_path)
            abs_paths.append(abs_path)
            rel_paths.append(relpath)
    return abs_paths, rel_paths


def get_language_from_ext(path) -> Tuple[str, str, bool]:
    root, ext = os.path.splitext(path)
    language_info = EXT_TO_LANGUAGE_DATA.get(ext, {})
    is_code = language_info.get("is_code", False)
    language = language_info.get("language_mode", None)
    lsp_language = LANGUAGE_TO_LSP_LANGUAGE_MAP.get(language, None)
    return lsp_language, language, is_code


def keep_file_for_language(root, file, language):
    file_language, _, is_code = get_language_from_ext(file)
    if file_language == language and is_code:
        return True
    return False
    # match language:
    #     case "javascript" | "typescript":
    #         if file.endswith("config.json") or file.endswith("package.json"):
    #             print("CallGraphBuilder :: keeping file: ", os.path.join(root, file))
    #             return True
    #     case _:
    #         return False


def copy_and_split_root_by_language_group(abs_root_path) -> List[LanguageGroup]:
    abs_paths, _ = get_all_paths_from_root_relative(abs_root_path)
    languages = set()

    for p in abs_paths:
        lsp_language, language, is_code = get_language_from_ext(p)
        if is_code:
            languages.add(lsp_language)
    languages = [lang for lang in languages if lang]

    copy_paths = []
    # copy the root directory into a temporary directory per language
    for _ in range(len(languages)):
        tmp_parent_dir = tempfile.mkdtemp(prefix="scope_")
        print(f"Copying {abs_root_path} to {tmp_parent_dir}")
        shutil.copytree(abs_root_path, tmp_parent_dir, dirs_exist_ok=True)
        copy_paths.append(tmp_parent_dir)

    for copy_path, language in zip(copy_paths, languages):
        for root, dirs, files in os.walk(copy_path):
            for file in files:
                if keep_file_for_language(root, file, language):
                    continue
                else:
                    # print(f"removing file: {os.path.join(root, file)}")
                    os.remove(os.path.join(root, file))

    # remove copy_paths that only have directories and no files
    nonempty_copy_paths = []
    for copy_path, language in zip(copy_paths, languages):
        files_set = set()
        for root, dirs, files in os.walk(copy_path):
            for file in files:
                files_set.add(file)
        if not files_set:
            # print(f"copy_path: {copy_path} is empty")
            shutil.rmtree(copy_path)
            continue

        # TODO: figure out if this screws up with path removal on cleanup
        child_dirs = os.listdir(copy_path)
        if len(child_dirs) == 1:
            copy_path = os.path.join(copy_path, child_dirs[0])
        nonempty_copy_paths.append(LanguageGroup(language, copy_path))

    return nonempty_copy_paths



EXT_TO_LANGUAGE_DATA = {
    ".bsl": {"is_code": True, "language_mode": "text"},
    ".os": {"is_code": True, "language_mode": "text"},
    ".2da": {"is_code": False, "language_mode": "text"},
    ".4dm": {"is_code": True, "language_mode": "text"},
    ".abap": {"is_code": True, "language_mode": "abap"},
    ".asddls": {"is_code": True, "language_mode": "text"},
    ".abnf": {"is_code": False, "language_mode": "text"},
    ".asc": {"is_code": False, "language_mode": "text"},
    ".ash": {"is_code": True, "language_mode": "c_cpp"},
    ".aidl": {"is_code": True, "language_mode": "text"},
    ".al": {"is_code": True, "language_mode": "perl"},
    ".ampl": {"is_code": True, "language_mode": "text"},
    ".mod": {"is_code": False, "language_mode": "xml"},
    ".g4": {"is_code": True, "language_mode": "text"},
    ".apib": {"is_code": False, "language_mode": "markdown"},
    ".apl": {"is_code": True, "language_mode": "text"},
    ".dyalog": {"is_code": True, "language_mode": "text"},
    ".asl": {"is_code": True, "language_mode": "text"},
    ".dsl": {"is_code": True, "language_mode": "text"},
    ".asn": {"is_code": False, "language_mode": "text"},
    ".asn1": {"is_code": False, "language_mode": "text"},
    ".asax": {"is_code": True, "language_mode": "text"},
    ".ascx": {"is_code": True, "language_mode": "text"},
    ".ashx": {"is_code": True, "language_mode": "text"},
    ".asmx": {"is_code": True, "language_mode": "text"},
    ".aspx": {"is_code": True, "language_mode": "text"},
    ".axd": {"is_code": True, "language_mode": "text"},
    ".dats": {"is_code": True, "language_mode": "ocaml"},
    ".hats": {"is_code": True, "language_mode": "ocaml"},
    ".sats": {"is_code": True, "language_mode": "ocaml"},
    ".as": {"is_code": True, "language_mode": "text"},
    ".adb": {"is_code": True, "language_mode": "ada"},
    ".ada": {"is_code": True, "language_mode": "ada"},
    ".ads": {"is_code": True, "language_mode": "ada"},
    ".txt": {"is_code": False, "language_mode": "text"},
    ".afm": {"is_code": False, "language_mode": "text"},
    ".agda": {"is_code": True, "language_mode": "text"},
    ".als": {"is_code": True, "language_mode": "text"},
    ".OutJob": {"is_code": False, "language_mode": "ini"},
    ".PcbDoc": {"is_code": False, "language_mode": "ini"},
    ".PrjPCB": {"is_code": False, "language_mode": "ini"},
    ".SchDoc": {"is_code": False, "language_mode": "ini"},
    ".angelscript": {"is_code": True, "language_mode": "text"},
    ".antlers.html": {"is_code": False, "language_mode": "text"},
    ".antlers.php": {"is_code": False, "language_mode": "text"},
    ".antlers.xml": {"is_code": False, "language_mode": "text"},
    ".apacheconf": {"is_code": False, "language_mode": "apache_conf"},
    ".vhost": {"is_code": False, "language_mode": "text"},
    ".cls": {"is_code": True, "language_mode": "text"},
    ".agc": {"is_code": True, "language_mode": "assembly_x86"},
    ".applescript": {"is_code": True, "language_mode": "applescript"},
    ".scpt": {"is_code": True, "language_mode": "applescript"},
    ".arc": {"is_code": True, "language_mode": "text"},
    ".asciidoc": {"is_code": False, "language_mode": "asciidoc"},
    ".adoc": {"is_code": False, "language_mode": "asciidoc"},
    ".aj": {"is_code": True, "language_mode": "text"},
    ".asm": {"is_code": True, "language_mode": "assembly_x86"},
    ".a51": {"is_code": True, "language_mode": "assembly_x86"},
    ".i": {"is_code": True, "language_mode": "c_cpp"},
    ".inc": {"is_code": True, "language_mode": "text"},
    ".nas": {"is_code": True, "language_mode": "text"},
    ".nasm": {"is_code": True, "language_mode": "assembly_x86"},
    ".astro": {"is_code": False, "language_mode": "html"},
    ".asy": {"is_code": False, "language_mode": "text"},
    ".aug": {"is_code": True, "language_mode": "text"},
    ".ahk": {"is_code": True, "language_mode": "autohotkey"},
    ".ahkl": {"is_code": True, "language_mode": "autohotkey"},
    ".au3": {"is_code": True, "language_mode": "autohotkey"},
    ".avdl": {"is_code": False, "language_mode": "text"},
    ".awk": {"is_code": True, "language_mode": "text"},
    ".auk": {"is_code": True, "language_mode": "text"},
    ".gawk": {"is_code": True, "language_mode": "text"},
    ".mawk": {"is_code": True, "language_mode": "text"},
    ".nawk": {"is_code": True, "language_mode": "text"},
    ".bas": {"is_code": True, "language_mode": "text"},
    ".bal": {"is_code": True, "language_mode": "text"},
    ".bat": {"is_code": True, "language_mode": "batchfile"},
    ".cmd": {"is_code": True, "language_mode": "batchfile"},
    ".bf": {"is_code": True, "language_mode": "text"},
    ".befunge": {"is_code": True, "language_mode": "text"},
    ".be": {"is_code": True, "language_mode": "text"},
    ".bib": {"is_code": False, "language_mode": "tex"},
    ".bibtex": {"is_code": False, "language_mode": "tex"},
    ".bicep": {"is_code": True, "language_mode": "text"},
    ".bs": {"is_code": True, "language_mode": "text"},
    ".bison": {"is_code": True, "language_mode": "text"},
    ".bb": {"is_code": True, "language_mode": "clojure"},
    ".blade": {"is_code": False, "language_mode": "text"},
    ".blade.php": {"is_code": False, "language_mode": "text"},
    ".decls": {"is_code": True, "language_mode": "text"},
    ".bmx": {"is_code": True, "language_mode": "text"},
    ".bsv": {"is_code": True, "language_mode": "verilog"},
    ".boo": {"is_code": True, "language_mode": "text"},
    ".bpl": {"is_code": True, "language_mode": "text"},
    ".b": {"is_code": True, "language_mode": "text"},
    ".brs": {"is_code": True, "language_mode": "text"},
    ".c": {"is_code": True, "language_mode": "c_cpp"},
    ".cats": {"is_code": True, "language_mode": "c_cpp"},
    ".h": {"is_code": True, "language_mode": "objectivec"},
    ".idc": {"is_code": True, "language_mode": "c_cpp"},
    ".cs": {"is_code": True, "language_mode": "text"},
    ".cake": {"is_code": True, "language_mode": "coffee"},
    ".csx": {"is_code": True, "language_mode": "csharp"},
    ".linq": {"is_code": True, "language_mode": "csharp"},
    ".cpp": {"is_code": True, "language_mode": "c_cpp"},
    ".c++": {"is_code": True, "language_mode": "c_cpp"},
    ".cc": {"is_code": True, "language_mode": "c_cpp"},
    ".cp": {"is_code": True, "language_mode": "pascal"},
    ".cppm": {"is_code": True, "language_mode": "c_cpp"},
    ".cxx": {"is_code": True, "language_mode": "c_cpp"},
    ".h++": {"is_code": True, "language_mode": "c_cpp"},
    ".hh": {"is_code": True, "language_mode": "php"},
    ".hpp": {"is_code": True, "language_mode": "c_cpp"},
    ".hxx": {"is_code": True, "language_mode": "c_cpp"},
    ".inl": {"is_code": True, "language_mode": "c_cpp"},
    ".ino": {"is_code": True, "language_mode": "c_cpp"},
    ".ipp": {"is_code": True, "language_mode": "c_cpp"},
    ".ixx": {"is_code": True, "language_mode": "c_cpp"},
    ".re": {"is_code": True, "language_mode": "rust"},
    ".tcc": {"is_code": True, "language_mode": "c_cpp"},
    ".tpp": {"is_code": True, "language_mode": "c_cpp"},
    ".c-objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".chs": {"is_code": True, "language_mode": "haskell"},
    ".cds": {"is_code": True, "language_mode": "text"},
    ".cil": {"is_code": False, "language_mode": "text"},
    ".clp": {"is_code": True, "language_mode": "text"},
    ".cmake": {"is_code": True, "language_mode": "text"},
    ".cmake.in": {"is_code": True, "language_mode": "text"},
    ".cob": {"is_code": True, "language_mode": "cobol"},
    ".cbl": {"is_code": True, "language_mode": "cobol"},
    ".ccp": {"is_code": True, "language_mode": "cobol"},
    ".cobol": {"is_code": True, "language_mode": "cobol"},
    ".cpy": {"is_code": True, "language_mode": "cobol"},
    ".dae": {"is_code": False, "language_mode": "xml"},
    ".cson": {"is_code": False, "language_mode": "coffee"},
    ".css": {"is_code": True, "language_mode": "css"},
    ".csv": {"is_code": False, "language_mode": "text"},
    ".cue": {"is_code": False, "language_mode": "text"},
    ".w": {"is_code": True, "language_mode": "text"},
    ".cabal": {"is_code": False, "language_mode": "haskell"},
    ".cdc": {"is_code": True, "language_mode": "text"},
    ".cairo": {"is_code": True, "language_mode": "text"},
    ".mligo": {"is_code": True, "language_mode": "ocaml"},
    ".capnp": {"is_code": True, "language_mode": "text"},
    ".mss": {"is_code": True, "language_mode": "text"},
    ".ceylon": {"is_code": True, "language_mode": "text"},
    ".chpl": {"is_code": True, "language_mode": "text"},
    ".ch": {"is_code": True, "language_mode": "text"},
    ".crc32": {"is_code": False, "language_mode": "text"},
    ".md2": {"is_code": True, "language_mode": "text"},
    ".md4": {"is_code": True, "language_mode": "text"},
    ".md5": {"is_code": True, "language_mode": "text"},
    ".sha1": {"is_code": False, "language_mode": "text"},
    ".sha2": {"is_code": False, "language_mode": "text"},
    ".sha224": {"is_code": False, "language_mode": "text"},
    ".sha256": {"is_code": False, "language_mode": "text"},
    ".sha256sum": {"is_code": False, "language_mode": "text"},
    ".sha3": {"is_code": False, "language_mode": "text"},
    ".sha384": {"is_code": False, "language_mode": "text"},
    ".sha512": {"is_code": False, "language_mode": "text"},
    ".ck": {"is_code": True, "language_mode": "java"},
    ".circom": {"is_code": True, "language_mode": "text"},
    ".cirru": {"is_code": True, "language_mode": "cirru"},
    ".clw": {"is_code": True, "language_mode": "text"},
    ".clar": {"is_code": True, "language_mode": "lisp"},
    ".asp": {"is_code": True, "language_mode": "text"},
    ".icl": {"is_code": True, "language_mode": "text"},
    ".dcl": {"is_code": True, "language_mode": "text"},
    ".click": {"is_code": True, "language_mode": "text"},
    ".clj": {"is_code": True, "language_mode": "clojure"},
    ".boot": {"is_code": True, "language_mode": "clojure"},
    ".cl2": {"is_code": True, "language_mode": "clojure"},
    ".cljc": {"is_code": True, "language_mode": "clojure"},
    ".cljs": {"is_code": True, "language_mode": "clojure"},
    ".cljs.hl": {"is_code": True, "language_mode": "clojure"},
    ".cljscm": {"is_code": True, "language_mode": "clojure"},
    ".cljx": {"is_code": True, "language_mode": "clojure"},
    ".hic": {"is_code": True, "language_mode": "clojure"},
    ".soy": {"is_code": False, "language_mode": "soy_template"},
    ".conllu": {"is_code": False, "language_mode": "text"},
    ".conll": {"is_code": False, "language_mode": "text"},
    ".ql": {"is_code": True, "language_mode": "text"},
    ".qll": {"is_code": True, "language_mode": "text"},
    ".coffee": {"is_code": True, "language_mode": "coffee"},
    "._coffee": {"is_code": True, "language_mode": "coffee"},
    ".cjsx": {"is_code": True, "language_mode": "coffee"},
    ".iced": {"is_code": True, "language_mode": "coffee"},
    ".cfm": {"is_code": True, "language_mode": "coldfusion"},
    ".cfml": {"is_code": True, "language_mode": "coldfusion"},
    ".cfc": {"is_code": True, "language_mode": "coldfusion"},
    ".lisp": {"is_code": True, "language_mode": "lisp"},
    ".asd": {"is_code": True, "language_mode": "lisp"},
    ".cl": {"is_code": True, "language_mode": "c_cpp"},
    ".l": {"is_code": False, "language_mode": "text"},
    ".lsp": {"is_code": True, "language_mode": "lisp"},
    ".ny": {"is_code": True, "language_mode": "lisp"},
    ".podsl": {"is_code": True, "language_mode": "lisp"},
    ".sexp": {"is_code": True, "language_mode": "lisp"},
    ".cwl": {"is_code": True, "language_mode": "yaml"},
    ".cps": {"is_code": True, "language_mode": "pascal"},
    ".coq": {"is_code": True, "language_mode": "text"},
    ".v": {"is_code": True, "language_mode": "verilog"},
    ".cppobjdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".c++-objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".c++objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".cpp-objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".cxx-objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".creole": {"is_code": False, "language_mode": "text"},
    ".cr": {"is_code": True, "language_mode": "ruby"},
    ".orc": {"is_code": True, "language_mode": "csound_orchestra"},
    ".udo": {"is_code": True, "language_mode": "csound_orchestra"},
    ".csd": {"is_code": True, "language_mode": "csound_document"},
    ".sco": {"is_code": True, "language_mode": "csound_score"},
    ".cu": {"is_code": True, "language_mode": "c_cpp"},
    ".cuh": {"is_code": True, "language_mode": "c_cpp"},
    ".curry": {"is_code": True, "language_mode": "haskell"},
    ".cy": {"is_code": True, "language_mode": "javascript"},
    ".cyp": {"is_code": True, "language_mode": "text"},
    ".cypher": {"is_code": True, "language_mode": "text"},
    ".pyx": {"is_code": True, "language_mode": "text"},
    ".pxd": {"is_code": True, "language_mode": "text"},
    ".pxi": {"is_code": True, "language_mode": "text"},
    ".d": {"is_code": True, "language_mode": "makefile"},
    ".di": {"is_code": True, "language_mode": "d"},
    ".d-objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".com": {"is_code": False, "language_mode": "text"},
    ".dm": {"is_code": True, "language_mode": "c_cpp"},
    ".zone": {"is_code": False, "language_mode": "text"},
    ".arpa": {"is_code": False, "language_mode": "text"},
    ".dfy": {"is_code": True, "language_mode": "text"},
    ".darcspatch": {"is_code": False, "language_mode": "text"},
    ".dpatch": {"is_code": False, "language_mode": "text"},
    ".dart": {"is_code": True, "language_mode": "dart"},
    ".dwl": {"is_code": True, "language_mode": "text"},
    ".dsc": {"is_code": True, "language_mode": "yaml"},
    ".dhall": {"is_code": True, "language_mode": "haskell"},
    ".diff": {"is_code": False, "language_mode": "diff"},
    ".patch": {"is_code": False, "language_mode": "diff"},
    ".x": {"is_code": True, "language_mode": "c_cpp"},
    ".dockerfile": {"is_code": True, "language_mode": "dockerfile"},
    ".djs": {"is_code": True, "language_mode": "text"},
    ".env": {"is_code": True, "language_mode": "text"},
    ".example": {"is_code": True, "language_mode": "text"},
    ".template": {"is_code": True, "language_mode": "text"},
    ".dylan": {"is_code": True, "language_mode": "text"},
    ".dyl": {"is_code": True, "language_mode": "text"},
    ".intr": {"is_code": True, "language_mode": "text"},
    ".lid": {"is_code": True, "language_mode": "text"},
    ".e": {"is_code": True, "language_mode": "text"},
    ".eml": {"is_code": False, "language_mode": "text"},
    ".mbox": {"is_code": False, "language_mode": "text"},
    ".ebnf": {"is_code": False, "language_mode": "text"},
    ".ecl": {"is_code": True, "language_mode": "prolog"},
    ".eclxml": {"is_code": True, "language_mode": "text"},
    ".ejs": {"is_code": False, "language_mode": "ejs"},
    ".ect": {"is_code": False, "language_mode": "ejs"},
    ".ejs.t": {"is_code": False, "language_mode": "ejs"},
    ".jst": {"is_code": False, "language_mode": "ejs"},
    ".eq": {"is_code": True, "language_mode": "csharp"},
    ".sch": {"is_code": False, "language_mode": "xml"},
    ".brd": {"is_code": False, "language_mode": "text"},
    ".eb": {"is_code": False, "language_mode": "python"},
    ".epj": {"is_code": False, "language_mode": "json"},
    ".html": {"is_code": True, "language_mode": "html"},
    ".editorconfig": {"is_code": False, "language_mode": "ini"},
    ".edc": {"is_code": False, "language_mode": "c_cpp"},
    ".ex": {"is_code": True, "language_mode": "text"},
    ".exs": {"is_code": True, "language_mode": "elixir"},
    ".elm": {"is_code": True, "language_mode": "elm"},
    ".elv": {"is_code": True, "language_mode": "text"},
    ".el": {"is_code": True, "language_mode": "lisp"},
    ".emacs": {"is_code": True, "language_mode": "lisp"},
    ".emacs.desktop": {"is_code": True, "language_mode": "lisp"},
    ".em": {"is_code": True, "language_mode": "coffee"},
    ".emberscript": {"is_code": True, "language_mode": "coffee"},
    ".erl": {"is_code": True, "language_mode": "erlang"},
    ".app": {"is_code": False, "language_mode": "erlang"},
    ".app.src": {"is_code": True, "language_mode": "erlang"},
    ".es": {"is_code": True, "language_mode": "javascript"},
    ".escript": {"is_code": True, "language_mode": "erlang"},
    ".hrl": {"is_code": True, "language_mode": "erlang"},
    ".xrl": {"is_code": True, "language_mode": "erlang"},
    ".yrl": {"is_code": True, "language_mode": "erlang"},
    ".fs": {"is_code": True, "language_mode": "glsl"},
    ".fsi": {"is_code": True, "language_mode": "text"},
    ".fsx": {"is_code": True, "language_mode": "text"},
    ".fst": {"is_code": True, "language_mode": "text"},
    ".fsti": {"is_code": True, "language_mode": "text"},
    ".flf": {"is_code": False, "language_mode": "text"},
    ".fx": {"is_code": True, "language_mode": "text"},
    ".flux": {"is_code": True, "language_mode": "text"},
    ".factor": {"is_code": True, "language_mode": "text"},
    ".fy": {"is_code": True, "language_mode": "text"},
    ".fancypack": {"is_code": True, "language_mode": "text"},
    ".fan": {"is_code": True, "language_mode": "text"},
    ".dsp": {"is_code": False, "language_mode": "text"},
    ".fnl": {"is_code": True, "language_mode": "text"},
    ".f": {"is_code": True, "language_mode": "text"},
    ".ftl": {"is_code": True, "language_mode": "ftl"},
    ".for": {"is_code": True, "language_mode": "text"},
    ".eam.fs": {"is_code": False, "language_mode": "text"},
    ".fth": {"is_code": True, "language_mode": "forth"},
    ".4th": {"is_code": True, "language_mode": "forth"},
    ".forth": {"is_code": True, "language_mode": "forth"},
    ".fr": {"is_code": False, "language_mode": "text"},
    ".frt": {"is_code": True, "language_mode": "forth"},
    ".f77": {"is_code": True, "language_mode": "text"},
    ".fpp": {"is_code": True, "language_mode": "text"},
    ".f90": {"is_code": True, "language_mode": "text"},
    ".f03": {"is_code": True, "language_mode": "text"},
    ".f08": {"is_code": True, "language_mode": "text"},
    ".f95": {"is_code": True, "language_mode": "text"},
    ".bi": {"is_code": True, "language_mode": "text"},
    ".fut": {"is_code": True, "language_mode": "text"},
    ".g": {"is_code": True, "language_mode": "text"},
    ".cnc": {"is_code": True, "language_mode": "gcode"},
    ".gco": {"is_code": True, "language_mode": "gcode"},
    ".gcode": {"is_code": True, "language_mode": "gcode"},
    ".gaml": {"is_code": True, "language_mode": "text"},
    ".gms": {"is_code": True, "language_mode": "text"},
    ".gap": {"is_code": True, "language_mode": "text"},
    ".gd": {"is_code": True, "language_mode": "text"},
    ".gi": {"is_code": True, "language_mode": "text"},
    ".tst": {"is_code": True, "language_mode": "text"},
    ".md": {"is_code": True, "language_mode": "markdown"},
    ".gdb": {"is_code": True, "language_mode": "text"},
    ".gdbinit": {"is_code": True, "language_mode": "text"},
    ".ged": {"is_code": False, "language_mode": "text"},
    ".glsl": {"is_code": True, "language_mode": "glsl"},
    ".fp": {"is_code": True, "language_mode": "glsl"},
    ".frag": {"is_code": True, "language_mode": "javascript"},
    ".frg": {"is_code": True, "language_mode": "glsl"},
    ".fsh": {"is_code": True, "language_mode": "glsl"},
    ".fshader": {"is_code": True, "language_mode": "glsl"},
    ".geo": {"is_code": True, "language_mode": "glsl"},
    ".geom": {"is_code": True, "language_mode": "glsl"},
    ".glslf": {"is_code": True, "language_mode": "glsl"},
    ".glslv": {"is_code": True, "language_mode": "glsl"},
    ".gs": {"is_code": True, "language_mode": "javascript"},
    ".gshader": {"is_code": True, "language_mode": "glsl"},
    ".rchit": {"is_code": True, "language_mode": "glsl"},
    ".rmiss": {"is_code": True, "language_mode": "glsl"},
    ".shader": {"is_code": True, "language_mode": "text"},
    ".tesc": {"is_code": True, "language_mode": "glsl"},
    ".tese": {"is_code": True, "language_mode": "glsl"},
    ".vert": {"is_code": True, "language_mode": "glsl"},
    ".vrx": {"is_code": True, "language_mode": "glsl"},
    ".vsh": {"is_code": True, "language_mode": "glsl"},
    ".vshader": {"is_code": True, "language_mode": "glsl"},
    ".gn": {"is_code": False, "language_mode": "python"},
    ".gni": {"is_code": False, "language_mode": "python"},
    ".gsc": {"is_code": True, "language_mode": "c_cpp"},
    ".csc": {"is_code": True, "language_mode": "c_cpp"},
    ".gsh": {"is_code": True, "language_mode": "c_cpp"},
    ".gml": {"is_code": False, "language_mode": "xml"},
    ".gmi": {"is_code": False, "language_mode": "text"},
    ".4gl": {"is_code": True, "language_mode": "text"},
    ".per": {"is_code": False, "language_mode": "text"},
    ".kid": {"is_code": True, "language_mode": "xml"},
    ".ebuild": {"is_code": True, "language_mode": "sh"},
    ".eclass": {"is_code": True, "language_mode": "sh"},
    ".gbr": {"is_code": False, "language_mode": "text"},
    ".cmp": {"is_code": False, "language_mode": "text"},
    ".gbl": {"is_code": False, "language_mode": "text"},
    ".gbo": {"is_code": False, "language_mode": "text"},
    ".gbp": {"is_code": False, "language_mode": "text"},
    ".gbs": {"is_code": False, "language_mode": "text"},
    ".gko": {"is_code": False, "language_mode": "text"},
    ".gpb": {"is_code": False, "language_mode": "text"},
    ".gpt": {"is_code": False, "language_mode": "text"},
    ".gtl": {"is_code": False, "language_mode": "text"},
    ".gto": {"is_code": False, "language_mode": "text"},
    ".gtp": {"is_code": False, "language_mode": "text"},
    ".gts": {"is_code": False, "language_mode": "text"},
    ".ncl": {"is_code": False, "language_mode": "xml"},
    ".sol": {"is_code": True, "language_mode": "text"},
    ".po": {"is_code": False, "language_mode": "text"},
    ".pot": {"is_code": False, "language_mode": "text"},
    ".feature": {"is_code": True, "language_mode": "text"},
    ".story": {"is_code": True, "language_mode": "text"},
    ".gitconfig": {"is_code": False, "language_mode": "ini"},
    ".gleam": {"is_code": True, "language_mode": "text"},
    ".glf": {"is_code": True, "language_mode": "tcl"},
    ".bdf": {"is_code": False, "language_mode": "text"},
    ".gp": {"is_code": True, "language_mode": "text"},
    ".gnu": {"is_code": True, "language_mode": "text"},
    ".gnuplot": {"is_code": True, "language_mode": "text"},
    ".p": {"is_code": True, "language_mode": "text"},
    ".plot": {"is_code": True, "language_mode": "text"},
    ".plt": {"is_code": True, "language_mode": "prolog"},
    ".go": {"is_code": True, "language_mode": "golang"},
    ".gdnlib": {"is_code": False, "language_mode": "text"},
    ".gdns": {"is_code": False, "language_mode": "text"},
    ".tres": {"is_code": False, "language_mode": "text"},
    ".tscn": {"is_code": False, "language_mode": "text"},
    ".golo": {"is_code": True, "language_mode": "text"},
    ".gst": {"is_code": False, "language_mode": "xml"},
    ".gsx": {"is_code": True, "language_mode": "text"},
    ".vark": {"is_code": True, "language_mode": "text"},
    ".grace": {"is_code": True, "language_mode": "text"},
    ".gradle": {"is_code": False, "language_mode": "text"},
    ".gf": {"is_code": True, "language_mode": "haskell"},
    ".graphql": {"is_code": False, "language_mode": "text"},
    ".gql": {"is_code": False, "language_mode": "text"},
    ".graphqls": {"is_code": False, "language_mode": "text"},
    ".dot": {"is_code": False, "language_mode": "text"},
    ".gv": {"is_code": False, "language_mode": "text"},
    ".groovy": {"is_code": True, "language_mode": "groovy"},
    ".grt": {"is_code": True, "language_mode": "groovy"},
    ".gtpl": {"is_code": True, "language_mode": "groovy"},
    ".gvy": {"is_code": True, "language_mode": "groovy"},
    ".gsp": {"is_code": True, "language_mode": "jsp"},
    ".cfg": {"is_code": False, "language_mode": "ini"},
    ".hcl": {"is_code": True, "language_mode": "ruby"},
    ".nomad": {"is_code": True, "language_mode": "ruby"},
    ".tf": {"is_code": True, "language_mode": "ruby"},
    ".tfvars": {"is_code": True, "language_mode": "ruby"},
    ".workflow": {"is_code": False, "language_mode": "xml"},
    ".hlsl": {"is_code": True, "language_mode": "text"},
    ".cginc": {"is_code": True, "language_mode": "text"},
    ".fxh": {"is_code": True, "language_mode": "text"},
    ".hlsli": {"is_code": True, "language_mode": "text"},
    ".hocon": {"is_code": False, "language_mode": "text"},
    ".hta": {"is_code": False, "language_mode": "html"},
    ".htm": {"is_code": False, "language_mode": "html"},
    ".html.hl": {"is_code": False, "language_mode": "html"},
    ".xht": {"is_code": False, "language_mode": "html"},
    ".xhtml": {"is_code": False, "language_mode": "html"},
    ".ecr": {"is_code": False, "language_mode": "text"},
    ".eex": {"is_code": False, "language_mode": "text"},
    ".html.heex": {"is_code": False, "language_mode": "text"},
    ".html.leex": {"is_code": False, "language_mode": "text"},
    ".erb": {"is_code": False, "language_mode": "text"},
    ".erb.deface": {"is_code": False, "language_mode": "text"},
    ".rhtml": {"is_code": False, "language_mode": "text"},
    ".phtml": {"is_code": False, "language_mode": "php"},
    ".cshtml": {"is_code": False, "language_mode": "razor"},
    ".razor": {"is_code": False, "language_mode": "razor"},
    ".http": {"is_code": False, "language_mode": "text"},
    ".hxml": {"is_code": False, "language_mode": "text"},
    ".hack": {"is_code": True, "language_mode": "php"},
    ".hhi": {"is_code": True, "language_mode": "php"},
    ".php": {"is_code": True, "language_mode": "php"},
    ".haml": {"is_code": False, "language_mode": "haml"},
    ".haml.deface": {"is_code": False, "language_mode": "haml"},
    ".handlebars": {"is_code": False, "language_mode": "handlebars"},
    ".hbs": {"is_code": False, "language_mode": "handlebars"},
    ".hb": {"is_code": True, "language_mode": "text"},
    ".hs": {"is_code": True, "language_mode": "haskell"},
    ".hs-boot": {"is_code": True, "language_mode": "haskell"},
    ".hsc": {"is_code": True, "language_mode": "haskell"},
    ".hx": {"is_code": True, "language_mode": "haxe"},
    ".hxsl": {"is_code": True, "language_mode": "haxe"},
    ".q": {"is_code": True, "language_mode": "text"},
    ".hql": {"is_code": True, "language_mode": "sql"},
    ".hc": {"is_code": True, "language_mode": "c_cpp"},
    ".hy": {"is_code": True, "language_mode": "text"},
    ".pro": {"is_code": True, "language_mode": "text"},
    ".dlm": {"is_code": True, "language_mode": "text"},
    ".ipf": {"is_code": True, "language_mode": "text"},
    ".ini": {"is_code": False, "language_mode": "ini"},
    ".cnf": {"is_code": False, "language_mode": "ini"},
    ".dof": {"is_code": False, "language_mode": "ini"},
    ".lektorproject": {"is_code": False, "language_mode": "ini"},
    ".prefs": {"is_code": False, "language_mode": "ini"},
    ".properties": {"is_code": False, "language_mode": "properties"},
    ".url": {"is_code": False, "language_mode": "ini"},
    ".irclog": {"is_code": False, "language_mode": "text"},
    ".weechatlog": {"is_code": False, "language_mode": "text"},
    ".idr": {"is_code": True, "language_mode": "text"},
    ".lidr": {"is_code": True, "language_mode": "text"},
    ".gitignore": {"is_code": False, "language_mode": "gitignore"},
    ".ijm": {"is_code": True, "language_mode": "text"},
    ".imba": {"is_code": True, "language_mode": "text"},
    ".ni": {"is_code": True, "language_mode": "text"},
    ".i7x": {"is_code": True, "language_mode": "text"},
    ".ink": {"is_code": True, "language_mode": "text"},
    ".iss": {"is_code": True, "language_mode": "text"},
    ".isl": {"is_code": True, "language_mode": "text"},
    ".io": {"is_code": True, "language_mode": "io"},
    ".ik": {"is_code": True, "language_mode": "text"},
    ".thy": {"is_code": True, "language_mode": "text"},
    ".ijs": {"is_code": True, "language_mode": "text"},
    ".jcl": {"is_code": True, "language_mode": "text"},
    ".flex": {"is_code": True, "language_mode": "text"},
    ".jflex": {"is_code": True, "language_mode": "text"},
    ".json": {"is_code": True, "language_mode": "json"},
    ".4DForm": {"is_code": False, "language_mode": "json"},
    ".4DProject": {"is_code": False, "language_mode": "json"},
    ".avsc": {"is_code": False, "language_mode": "json"},
    ".geojson": {"is_code": False, "language_mode": "json"},
    ".gltf": {"is_code": False, "language_mode": "json"},
    ".har": {"is_code": False, "language_mode": "json"},
    ".ice": {"is_code": True, "language_mode": "text"},
    ".JSON-tmLanguage": {"is_code": False, "language_mode": "json"},
    ".jsonl": {"is_code": False, "language_mode": "json"},
    ".mcmeta": {"is_code": False, "language_mode": "json"},
    ".tfstate": {"is_code": False, "language_mode": "json"},
    ".tfstate.backup": {"is_code": False, "language_mode": "json"},
    ".topojson": {"is_code": False, "language_mode": "json"},
    ".webapp": {"is_code": False, "language_mode": "json"},
    ".webmanifest": {"is_code": False, "language_mode": "json"},
    ".yy": {"is_code": True, "language_mode": "text"},
    ".yyp": {"is_code": False, "language_mode": "json"},
    ".jsonc": {"is_code": False, "language_mode": "javascript"},
    ".code-snippets": {"is_code": False, "language_mode": "javascript"},
    ".sublime-build": {"is_code": False, "language_mode": "javascript"},
    ".sublime-commands": {"is_code": False, "language_mode": "javascript"},
    ".sublime-completions": {"is_code": False, "language_mode": "javascript"},
    ".sublime-keymap": {"is_code": False, "language_mode": "javascript"},
    ".sublime-macro": {"is_code": False, "language_mode": "javascript"},
    ".sublime-menu": {"is_code": False, "language_mode": "javascript"},
    ".sublime-mousemap": {"is_code": False, "language_mode": "javascript"},
    ".sublime-project": {"is_code": False, "language_mode": "javascript"},
    ".sublime-settings": {"is_code": False, "language_mode": "javascript"},
    ".sublime-theme": {"is_code": False, "language_mode": "javascript"},
    ".sublime-workspace": {"is_code": False, "language_mode": "javascript"},
    ".sublime_metrics": {"is_code": False, "language_mode": "javascript"},
    ".sublime_session": {"is_code": False, "language_mode": "javascript"},
    ".json5": {"is_code": False, "language_mode": "javascript"},
    ".jsonld": {"is_code": False, "language_mode": "javascript"},
    ".jq": {"is_code": True, "language_mode": "text"},
    ".janet": {"is_code": True, "language_mode": "scheme"},
    ".j": {"is_code": True, "language_mode": "text"},
    ".java": {"is_code": True, "language_mode": "java"},
    ".jav": {"is_code": True, "language_mode": "java"},
    ".jsh": {"is_code": True, "language_mode": "java"},
    ".jsp": {"is_code": True, "language_mode": "jsp"},
    ".tag": {"is_code": True, "language_mode": "jsp"},
    ".js": {"is_code": True, "language_mode": "javascript"},
    "._js": {"is_code": True, "language_mode": "javascript"},
    ".bones": {"is_code": True, "language_mode": "javascript"},
    ".cjs": {"is_code": True, "language_mode": "javascript"},
    ".es6": {"is_code": True, "language_mode": "javascript"},
    ".jake": {"is_code": True, "language_mode": "javascript"},
    ".javascript": {"is_code": True, "language_mode": "javascript"},
    ".jsb": {"is_code": True, "language_mode": "javascript"},
    ".jscad": {"is_code": True, "language_mode": "javascript"},
    ".jsfl": {"is_code": True, "language_mode": "javascript"},
    ".jslib": {"is_code": True, "language_mode": "javascript"},
    ".jsm": {"is_code": True, "language_mode": "javascript"},
    ".jspre": {"is_code": True, "language_mode": "javascript"},
    ".jss": {"is_code": True, "language_mode": "javascript"},
    ".jsx": {"is_code": True, "language_mode": "javascript"},
    ".mjs": {"is_code": True, "language_mode": "javascript"},
    ".njs": {"is_code": True, "language_mode": "javascript"},
    ".pac": {"is_code": True, "language_mode": "javascript"},
    ".sjs": {"is_code": True, "language_mode": "javascript"},
    ".ssjs": {"is_code": True, "language_mode": "javascript"},
    ".xsjs": {"is_code": True, "language_mode": "javascript"},
    ".xsjslib": {"is_code": True, "language_mode": "javascript"},
    ".js.erb": {"is_code": True, "language_mode": "javascript"},
    ".snap": {"is_code": False, "language_mode": "javascript"},
    ".mps": {"is_code": True, "language_mode": "xml"},
    ".mpl": {"is_code": True, "language_mode": "xml"},
    ".msd": {"is_code": True, "language_mode": "xml"},
    ".jinja": {"is_code": False, "language_mode": "django"},
    ".j2": {"is_code": False, "language_mode": "django"},
    ".jinja2": {"is_code": False, "language_mode": "django"},
    ".jison": {"is_code": True, "language_mode": "text"},
    ".jisonlex": {"is_code": True, "language_mode": "text"},
    ".ol": {"is_code": True, "language_mode": "text"},
    ".iol": {"is_code": True, "language_mode": "text"},
    ".jsonnet": {"is_code": True, "language_mode": "text"},
    ".libsonnet": {"is_code": True, "language_mode": "text"},
    ".jl": {"is_code": True, "language_mode": "julia"},
    ".ipynb": {"is_code": False, "language_mode": "json"},
    ".krl": {"is_code": True, "language_mode": "text"},
    ".ksy": {"is_code": True, "language_mode": "yaml"},
    ".kak": {"is_code": True, "language_mode": "text"},
    ".ks": {"is_code": False, "language_mode": "text"},
    ".kicad_pcb": {"is_code": False, "language_mode": "lisp"},
    ".kicad_mod": {"is_code": False, "language_mode": "lisp"},
    ".kicad_wks": {"is_code": False, "language_mode": "lisp"},
    ".kicad_sch": {"is_code": False, "language_mode": "text"},
    ".kit": {"is_code": False, "language_mode": "html"},
    ".kt": {"is_code": True, "language_mode": "text"},
    ".ktm": {"is_code": True, "language_mode": "text"},
    ".kts": {"is_code": True, "language_mode": "text"},
    ".csl": {"is_code": False, "language_mode": "xml"},
    ".kql": {"is_code": False, "language_mode": "text"},
    ".lfe": {"is_code": True, "language_mode": "lisp"},
    ".ll": {"is_code": True, "language_mode": "text"},
    ".lol": {"is_code": True, "language_mode": "text"},
    ".lsl": {"is_code": True, "language_mode": "lsl"},
    ".lslp": {"is_code": True, "language_mode": "lsl"},
    ".lvproj": {"is_code": True, "language_mode": "xml"},
    ".lvclass": {"is_code": True, "language_mode": "xml"},
    ".lvlib": {"is_code": True, "language_mode": "xml"},
    ".lark": {"is_code": False, "language_mode": "text"},
    ".lasso": {"is_code": True, "language_mode": "text"},
    ".las": {"is_code": True, "language_mode": "text"},
    ".lasso8": {"is_code": True, "language_mode": "text"},
    ".lasso9": {"is_code": True, "language_mode": "text"},
    ".latte": {"is_code": False, "language_mode": "smarty"},
    ".lean": {"is_code": True, "language_mode": "text"},
    ".hlean": {"is_code": True, "language_mode": "text"},
    ".less": {"is_code": False, "language_mode": "less"},
    ".lex": {"is_code": True, "language_mode": "text"},
    ".ligo": {"is_code": True, "language_mode": "pascal"},
    ".ly": {"is_code": True, "language_mode": "text"},
    ".ily": {"is_code": True, "language_mode": "text"},
    ".m": {"is_code": True, "language_mode": "objectivec"},
    ".ld": {"is_code": False, "language_mode": "text"},
    ".lds": {"is_code": False, "language_mode": "text"},
    ".liquid": {"is_code": False, "language_mode": "liquid"},
    ".lagda": {"is_code": True, "language_mode": "text"},
    ".litcoffee": {"is_code": True, "language_mode": "text"},
    ".coffee.md": {"is_code": True, "language_mode": "text"},
    ".lhs": {"is_code": True, "language_mode": "text"},
    ".ls": {"is_code": True, "language_mode": "text"},
    "._ls": {"is_code": True, "language_mode": "livescript"},
    ".xm": {"is_code": True, "language_mode": "text"},
    ".xi": {"is_code": True, "language_mode": "text"},
    ".lgt": {"is_code": True, "language_mode": "text"},
    ".logtalk": {"is_code": True, "language_mode": "text"},
    ".lookml": {"is_code": True, "language_mode": "yaml"},
    ".model.lkml": {"is_code": True, "language_mode": "yaml"},
    ".view.lkml": {"is_code": True, "language_mode": "yaml"},
    ".lua": {"is_code": True, "language_mode": "lua"},
    ".fcgi": {"is_code": True, "language_mode": "sh"},
    ".nse": {"is_code": True, "language_mode": "lua"},
    ".p8": {"is_code": True, "language_mode": "lua"},
    ".pd_lua": {"is_code": True, "language_mode": "lua"},
    ".rbxs": {"is_code": True, "language_mode": "lua"},
    ".rockspec": {"is_code": True, "language_mode": "lua"},
    ".wlua": {"is_code": True, "language_mode": "lua"},
    ".mumps": {"is_code": True, "language_mode": "text"},
    ".m4": {"is_code": True, "language_mode": "text"},
    ".mc": {"is_code": False, "language_mode": "ini"},
    ".matlab": {"is_code": True, "language_mode": "matlab"},
    ".ms": {"is_code": True, "language_mode": "assembly_x86"},
    ".mcr": {"is_code": True, "language_mode": "text"},
    ".mlir": {"is_code": True, "language_mode": "text"},
    ".mq4": {"is_code": True, "language_mode": "c_cpp"},
    ".mqh": {"is_code": True, "language_mode": "c_cpp"},
    ".mq5": {"is_code": True, "language_mode": "c_cpp"},
    ".mtml": {"is_code": False, "language_mode": "html"},
    ".muf": {"is_code": True, "language_mode": "forth"},
    ".m2": {"is_code": True, "language_mode": "text"},
    ".mak": {"is_code": True, "language_mode": "makefile"},
    ".make": {"is_code": True, "language_mode": "makefile"},
    ".makefile": {"is_code": True, "language_mode": "makefile"},
    ".mk": {"is_code": True, "language_mode": "makefile"},
    ".mkfile": {"is_code": True, "language_mode": "makefile"},
    ".mako": {"is_code": True, "language_mode": "text"},
    ".mao": {"is_code": True, "language_mode": "text"},
    ".livemd": {"is_code": False, "language_mode": "markdown"},
    ".markdown": {"is_code": False, "language_mode": "markdown"},
    ".mdown": {"is_code": False, "language_mode": "markdown"},
    ".mdwn": {"is_code": False, "language_mode": "markdown"},
    ".mdx": {"is_code": False, "language_mode": "markdown"},
    ".mkd": {"is_code": False, "language_mode": "markdown"},
    ".mkdn": {"is_code": False, "language_mode": "markdown"},
    ".mkdown": {"is_code": False, "language_mode": "markdown"},
    ".ronn": {"is_code": False, "language_mode": "markdown"},
    ".scd": {"is_code": True, "language_mode": "text"},
    ".workbook": {"is_code": False, "language_mode": "markdown"},
    ".marko": {"is_code": False, "language_mode": "text"},
    ".mask": {"is_code": False, "language_mode": "yaml"},
    ".mathematica": {"is_code": True, "language_mode": "text"},
    ".cdf": {"is_code": True, "language_mode": "text"},
    ".ma": {"is_code": True, "language_mode": "text"},
    ".mt": {"is_code": True, "language_mode": "text"},
    ".nb": {"is_code": False, "language_mode": "text"},
    ".nbp": {"is_code": True, "language_mode": "text"},
    ".wl": {"is_code": True, "language_mode": "text"},
    ".wlt": {"is_code": True, "language_mode": "text"},
    ".maxpat": {"is_code": True, "language_mode": "json"},
    ".maxhelp": {"is_code": True, "language_mode": "json"},
    ".maxproj": {"is_code": True, "language_mode": "json"},
    ".mxt": {"is_code": True, "language_mode": "json"},
    ".pat": {"is_code": True, "language_mode": "json"},
    ".moo": {"is_code": True, "language_mode": "text"},
    ".mmd": {"is_code": False, "language_mode": "text"},
    ".mermaid": {"is_code": False, "language_mode": "text"},
    ".metal": {"is_code": True, "language_mode": "c_cpp"},
    ".sln": {"is_code": False, "language_mode": "text"},
    ".minid": {"is_code": True, "language_mode": "text"},
    ".yaml": {"is_code": True, "language_mode": "yaml"},
    ".yml": {"is_code": True, "language_mode": "yaml"},
    ".mint": {"is_code": True, "language_mode": "text"},
    ".druby": {"is_code": True, "language_mode": "ruby"},
    ".duby": {"is_code": True, "language_mode": "ruby"},
    ".mirah": {"is_code": True, "language_mode": "ruby"},
    ".mo": {"is_code": True, "language_mode": "text"},
    ".i3": {"is_code": True, "language_mode": "text"},
    ".ig": {"is_code": True, "language_mode": "text"},
    ".m3": {"is_code": True, "language_mode": "text"},
    ".mg": {"is_code": True, "language_mode": "text"},
    ".mms": {"is_code": True, "language_mode": "text"},
    ".mmk": {"is_code": True, "language_mode": "text"},
    ".monkey": {"is_code": True, "language_mode": "text"},
    ".monkey2": {"is_code": True, "language_mode": "text"},
    ".moon": {"is_code": True, "language_mode": "text"},
    ".s": {"is_code": True, "language_mode": "assembly_x86"},
    ".x68": {"is_code": True, "language_mode": "assembly_x86"},
    ".move": {"is_code": True, "language_mode": "text"},
    ".muse": {"is_code": False, "language_mode": "text"},
    ".mustache": {"is_code": False, "language_mode": "smarty"},
    ".myt": {"is_code": True, "language_mode": "text"},
    ".nasl": {"is_code": True, "language_mode": "text"},
    ".neon": {"is_code": False, "language_mode": "text"},
    ".nl": {"is_code": True, "language_mode": "lisp"},
    ".nsi": {"is_code": True, "language_mode": "text"},
    ".nsh": {"is_code": True, "language_mode": "text"},
    ".nss": {"is_code": True, "language_mode": "c_cpp"},
    ".ne": {"is_code": True, "language_mode": "text"},
    ".nearley": {"is_code": True, "language_mode": "text"},
    ".n": {"is_code": False, "language_mode": "text"},
    ".axs": {"is_code": True, "language_mode": "text"},
    ".axi": {"is_code": True, "language_mode": "text"},
    ".axs.erb": {"is_code": True, "language_mode": "text"},
    ".axi.erb": {"is_code": True, "language_mode": "text"},
    ".nlogo": {"is_code": True, "language_mode": "lisp"},
    ".nf": {"is_code": True, "language_mode": "groovy"},
    ".nginx": {"is_code": False, "language_mode": "text"},
    ".nginxconf": {"is_code": False, "language_mode": "text"},
    ".nim": {"is_code": True, "language_mode": "text"},
    ".nim.cfg": {"is_code": True, "language_mode": "text"},
    ".nimble": {"is_code": True, "language_mode": "text"},
    ".nimrod": {"is_code": True, "language_mode": "text"},
    ".nims": {"is_code": True, "language_mode": "text"},
    ".ninja": {"is_code": False, "language_mode": "text"},
    ".nit": {"is_code": True, "language_mode": "text"},
    ".nix": {"is_code": True, "language_mode": "nix"},
    ".nu": {"is_code": True, "language_mode": "scheme"},
    ".numpy": {"is_code": True, "language_mode": "text"},
    ".numpyw": {"is_code": True, "language_mode": "text"},
    ".numsc": {"is_code": True, "language_mode": "text"},
    ".njk": {"is_code": False, "language_mode": "nunjucks"},
    ".ml": {"is_code": True, "language_mode": "text"},
    ".eliom": {"is_code": True, "language_mode": "ocaml"},
    ".eliomi": {"is_code": True, "language_mode": "ocaml"},
    ".ml4": {"is_code": True, "language_mode": "ocaml"},
    ".mli": {"is_code": True, "language_mode": "ocaml"},
    ".mll": {"is_code": True, "language_mode": "ocaml"},
    ".mly": {"is_code": True, "language_mode": "ocaml"},
    ".objdump": {"is_code": False, "language_mode": "assembly_x86"},
    ".odin": {"is_code": True, "language_mode": "text"},
    ".mm": {"is_code": False, "language_mode": "xml"},
    ".sj": {"is_code": True, "language_mode": "text"},
    ".omgrofl": {"is_code": True, "language_mode": "text"},
    ".opa": {"is_code": True, "language_mode": "text"},
    ".opal": {"is_code": True, "language_mode": "text"},
    ".rego": {"is_code": True, "language_mode": "text"},
    ".opencl": {"is_code": True, "language_mode": "c_cpp"},
    ".qasm": {"is_code": True, "language_mode": "text"},
    ".scad": {"is_code": True, "language_mode": "scad"},
    ".plist": {"is_code": False, "language_mode": "xml"},
    ".glyphs": {"is_code": False, "language_mode": "text"},
    ".fea": {"is_code": False, "language_mode": "text"},
    ".org": {"is_code": False, "language_mode": "text"},
    ".ox": {"is_code": True, "language_mode": "text"},
    ".oxh": {"is_code": True, "language_mode": "text"},
    ".oxo": {"is_code": True, "language_mode": "text"},
    ".oxygene": {"is_code": True, "language_mode": "text"},
    ".oz": {"is_code": True, "language_mode": "text"},
    ".p4": {"is_code": True, "language_mode": "text"},
    ".pddl": {"is_code": True, "language_mode": "text"},
    ".pegjs": {"is_code": True, "language_mode": "javascript"},
    ".aw": {"is_code": True, "language_mode": "php"},
    ".ctp": {"is_code": True, "language_mode": "php"},
    ".php3": {"is_code": True, "language_mode": "php"},
    ".php4": {"is_code": True, "language_mode": "php"},
    ".php5": {"is_code": True, "language_mode": "php"},
    ".phps": {"is_code": True, "language_mode": "php"},
    ".phpt": {"is_code": True, "language_mode": "php"},
    ".pls": {"is_code": True, "language_mode": "sql"},
    ".bdy": {"is_code": True, "language_mode": "sql"},
    ".ddl": {"is_code": False, "language_mode": "sql"},
    ".fnc": {"is_code": True, "language_mode": "sql"},
    ".pck": {"is_code": True, "language_mode": "sql"},
    ".pkb": {"is_code": True, "language_mode": "sql"},
    ".pks": {"is_code": True, "language_mode": "sql"},
    ".plb": {"is_code": True, "language_mode": "sql"},
    ".plsql": {"is_code": True, "language_mode": "sql"},
    ".prc": {"is_code": False, "language_mode": "sql"},
    ".spc": {"is_code": True, "language_mode": "sql"},
    ".sql": {"is_code": True, "language_mode": "sql"},
    ".tpb": {"is_code": True, "language_mode": "sql"},
    ".tps": {"is_code": True, "language_mode": "sql"},
    ".trg": {"is_code": True, "language_mode": "sql"},
    ".vw": {"is_code": True, "language_mode": "sql"},
    ".pgsql": {"is_code": True, "language_mode": "pgsql"},
    ".pov": {"is_code": True, "language_mode": "text"},
    ".pan": {"is_code": True, "language_mode": "text"},
    ".psc": {"is_code": True, "language_mode": "text"},
    ".parrot": {"is_code": True, "language_mode": "text"},
    ".pasm": {"is_code": True, "language_mode": "text"},
    ".pir": {"is_code": True, "language_mode": "text"},
    ".pas": {"is_code": True, "language_mode": "pascal"},
    ".dfm": {"is_code": True, "language_mode": "pascal"},
    ".dpr": {"is_code": True, "language_mode": "pascal"},
    ".lpr": {"is_code": True, "language_mode": "pascal"},
    ".pascal": {"is_code": True, "language_mode": "pascal"},
    ".pp": {"is_code": True, "language_mode": "text"},
    ".pwn": {"is_code": True, "language_mode": "text"},
    ".sma": {"is_code": True, "language_mode": "text"},
    ".pep": {"is_code": True, "language_mode": "text"},
    ".pl": {"is_code": True, "language_mode": "perl"},
    ".cgi": {"is_code": True, "language_mode": "sh"},
    ".perl": {"is_code": True, "language_mode": "perl"},
    ".ph": {"is_code": True, "language_mode": "perl"},
    ".plx": {"is_code": True, "language_mode": "perl"},
    ".pm": {"is_code": False, "language_mode": "c_cpp"},
    ".psgi": {"is_code": True, "language_mode": "perl"},
    ".t": {"is_code": True, "language_mode": "text"},
    ".pic": {"is_code": False, "language_mode": "text"},
    ".chem": {"is_code": False, "language_mode": "text"},
    ".pkl": {"is_code": False, "language_mode": "text"},
    ".pig": {"is_code": True, "language_mode": "text"},
    ".pike": {"is_code": True, "language_mode": "text"},
    ".pmod": {"is_code": True, "language_mode": "text"},
    ".puml": {"is_code": False, "language_mode": "text"},
    ".iuml": {"is_code": False, "language_mode": "text"},
    ".plantuml": {"is_code": False, "language_mode": "text"},
    ".pod": {"is_code": False, "language_mode": "perl"},
    ".pod6": {"is_code": False, "language_mode": "perl"},
    ".pogo": {"is_code": True, "language_mode": "text"},
    ".polar": {"is_code": True, "language_mode": "text"},
    ".pony": {"is_code": True, "language_mode": "text"},
    ".por": {"is_code": True, "language_mode": "text"},
    ".pcss": {"is_code": False, "language_mode": "text"},
    ".postcss": {"is_code": False, "language_mode": "text"},
    ".ps": {"is_code": False, "language_mode": "text"},
    ".eps": {"is_code": False, "language_mode": "text"},
    ".epsi": {"is_code": False, "language_mode": "text"},
    ".pfa": {"is_code": False, "language_mode": "text"},
    ".pbt": {"is_code": False, "language_mode": "text"},
    ".sra": {"is_code": True, "language_mode": "text"},
    ".sru": {"is_code": True, "language_mode": "text"},
    ".srw": {"is_code": True, "language_mode": "text"},
    ".ps1": {"is_code": True, "language_mode": "powershell"},
    ".psd1": {"is_code": True, "language_mode": "powershell"},
    ".psm1": {"is_code": True, "language_mode": "powershell"},
    ".prisma": {"is_code": False, "language_mode": "text"},
    ".pde": {"is_code": True, "language_mode": "text"},
    ".prolog": {"is_code": True, "language_mode": "prolog"},
    ".yap": {"is_code": True, "language_mode": "prolog"},
    ".pml": {"is_code": True, "language_mode": "text"},
    ".spin": {"is_code": True, "language_mode": "text"},
    ".proto": {"is_code": False, "language_mode": "protobuf"},
    ".textproto": {"is_code": False, "language_mode": "text"},
    ".pbtxt": {"is_code": False, "language_mode": "text"},
    ".pub": {"is_code": False, "language_mode": "text"},
    ".jade": {"is_code": False, "language_mode": "jade"},
    ".pug": {"is_code": False, "language_mode": "jade"},
    ".pd": {"is_code": False, "language_mode": "text"},
    ".pb": {"is_code": True, "language_mode": "text"},
    ".pbi": {"is_code": True, "language_mode": "text"},
    ".purs": {"is_code": True, "language_mode": "haskell"},
    ".arr": {"is_code": True, "language_mode": "python"},
    ".py": {"is_code": True, "language_mode": "python"},
    ".gyp": {"is_code": True, "language_mode": "python"},
    ".gypi": {"is_code": True, "language_mode": "python"},
    ".lmi": {"is_code": True, "language_mode": "python"},
    ".py3": {"is_code": True, "language_mode": "python"},
    ".pyde": {"is_code": True, "language_mode": "python"},
    ".pyi": {"is_code": True, "language_mode": "python"},
    ".pyp": {"is_code": True, "language_mode": "python"},
    ".pyt": {"is_code": True, "language_mode": "python"},
    ".pyw": {"is_code": True, "language_mode": "python"},
    ".rpy": {"is_code": True, "language_mode": "python"},
    ".spec": {"is_code": True, "language_mode": "ruby"},
    ".tac": {"is_code": True, "language_mode": "python"},
    ".wsgi": {"is_code": True, "language_mode": "python"},
    ".xpy": {"is_code": True, "language_mode": "python"},
    ".pytb": {"is_code": False, "language_mode": "text"},
    ".qs": {"is_code": True, "language_mode": "javascript"},
    ".qml": {"is_code": True, "language_mode": "text"},
    ".qbs": {"is_code": True, "language_mode": "text"},
    ".pri": {"is_code": True, "language_mode": "text"},
    ".r": {"is_code": True, "language_mode": "text"},
    ".rd": {"is_code": True, "language_mode": "r"},
    ".rsx": {"is_code": True, "language_mode": "r"},
    ".raml": {"is_code": False, "language_mode": "yaml"},
    ".rdoc": {"is_code": False, "language_mode": "rdoc"},
    ".rbbas": {"is_code": True, "language_mode": "text"},
    ".rbfrm": {"is_code": True, "language_mode": "text"},
    ".rbmnu": {"is_code": True, "language_mode": "text"},
    ".rbres": {"is_code": True, "language_mode": "text"},
    ".rbtbar": {"is_code": True, "language_mode": "text"},
    ".rbuistate": {"is_code": True, "language_mode": "text"},
    ".rexx": {"is_code": True, "language_mode": "text"},
    ".pprx": {"is_code": True, "language_mode": "text"},
    ".rex": {"is_code": True, "language_mode": "text"},
    ".qmd": {"is_code": False, "language_mode": "markdown"},
    ".rmd": {"is_code": False, "language_mode": "markdown"},
    ".rpgle": {"is_code": True, "language_mode": "text"},
    ".sqlrpgle": {"is_code": True, "language_mode": "text"},
    ".rnh": {"is_code": False, "language_mode": "text"},
    ".rno": {"is_code": False, "language_mode": "text"},
    ".rkt": {"is_code": True, "language_mode": "lisp"},
    ".rktd": {"is_code": True, "language_mode": "lisp"},
    ".rktl": {"is_code": True, "language_mode": "lisp"},
    ".scrbl": {"is_code": True, "language_mode": "lisp"},
    ".rl": {"is_code": True, "language_mode": "text"},
    ".6pl": {"is_code": True, "language_mode": "perl"},
    ".6pm": {"is_code": True, "language_mode": "perl"},
    ".nqp": {"is_code": True, "language_mode": "perl"},
    ".p6": {"is_code": True, "language_mode": "perl"},
    ".p6l": {"is_code": True, "language_mode": "perl"},
    ".p6m": {"is_code": True, "language_mode": "perl"},
    ".pl6": {"is_code": True, "language_mode": "perl"},
    ".pm6": {"is_code": True, "language_mode": "perl"},
    ".raku": {"is_code": True, "language_mode": "perl"},
    ".rakumod": {"is_code": True, "language_mode": "perl"},
    ".rsc": {"is_code": True, "language_mode": "text"},
    ".raw": {"is_code": False, "language_mode": "text"},
    ".res": {"is_code": False, "language_mode": "xml"},
    ".rei": {"is_code": True, "language_mode": "rust"},
    ".religo": {"is_code": True, "language_mode": "rust"},
    ".reb": {"is_code": True, "language_mode": "text"},
    ".r2": {"is_code": True, "language_mode": "text"},
    ".r3": {"is_code": True, "language_mode": "text"},
    ".rebol": {"is_code": True, "language_mode": "text"},
    ".red": {"is_code": True, "language_mode": "text"},
    ".reds": {"is_code": True, "language_mode": "text"},
    ".cw": {"is_code": True, "language_mode": "text"},
    ".regexp": {"is_code": False, "language_mode": "text"},
    ".regex": {"is_code": False, "language_mode": "text"},
    ".rs": {"is_code": True, "language_mode": "rust"},
    ".rsh": {"is_code": True, "language_mode": "text"},
    ".rtf": {"is_code": False, "language_mode": "text"},
    ".ring": {"is_code": True, "language_mode": "text"},
    ".riot": {"is_code": False, "language_mode": "html"},
    ".robot": {"is_code": True, "language_mode": "text"},
    ".roff": {"is_code": False, "language_mode": "text"},
    ".1": {"is_code": False, "language_mode": "text"},
    ".1in": {"is_code": False, "language_mode": "text"},
    ".1m": {"is_code": False, "language_mode": "text"},
    ".1x": {"is_code": False, "language_mode": "text"},
    ".2": {"is_code": False, "language_mode": "text"},
    ".3": {"is_code": False, "language_mode": "text"},
    ".3in": {"is_code": False, "language_mode": "text"},
    ".3m": {"is_code": False, "language_mode": "text"},
    ".3p": {"is_code": False, "language_mode": "text"},
    ".3pm": {"is_code": False, "language_mode": "text"},
    ".3qt": {"is_code": False, "language_mode": "text"},
    ".3x": {"is_code": False, "language_mode": "text"},
    ".4": {"is_code": False, "language_mode": "text"},
    ".5": {"is_code": False, "language_mode": "text"},
    ".6": {"is_code": False, "language_mode": "text"},
    ".7": {"is_code": False, "language_mode": "text"},
    ".8": {"is_code": False, "language_mode": "text"},
    ".9": {"is_code": False, "language_mode": "text"},
    ".man": {"is_code": False, "language_mode": "text"},
    ".mdoc": {"is_code": False, "language_mode": "text"},
    ".me": {"is_code": False, "language_mode": "text"},
    ".nr": {"is_code": False, "language_mode": "text"},
    ".tmac": {"is_code": False, "language_mode": "text"},
    ".rg": {"is_code": True, "language_mode": "clojure"},
    ".rb": {"is_code": True, "language_mode": "ruby"},
    ".builder": {"is_code": True, "language_mode": "ruby"},
    ".eye": {"is_code": True, "language_mode": "ruby"},
    ".gemspec": {"is_code": True, "language_mode": "ruby"},
    ".god": {"is_code": True, "language_mode": "ruby"},
    ".jbuilder": {"is_code": True, "language_mode": "ruby"},
    ".mspec": {"is_code": True, "language_mode": "ruby"},
    ".pluginspec": {"is_code": False, "language_mode": "xml"},
    ".podspec": {"is_code": True, "language_mode": "ruby"},
    ".prawn": {"is_code": True, "language_mode": "ruby"},
    ".rabl": {"is_code": True, "language_mode": "ruby"},
    ".rake": {"is_code": True, "language_mode": "ruby"},
    ".rbi": {"is_code": True, "language_mode": "ruby"},
    ".rbuild": {"is_code": True, "language_mode": "ruby"},
    ".rbw": {"is_code": True, "language_mode": "ruby"},
    ".rbx": {"is_code": True, "language_mode": "ruby"},
    ".ru": {"is_code": True, "language_mode": "ruby"},
    ".ruby": {"is_code": True, "language_mode": "ruby"},
    ".thor": {"is_code": True, "language_mode": "ruby"},
    ".watchr": {"is_code": True, "language_mode": "ruby"},
    ".rs.in": {"is_code": True, "language_mode": "rust"},
    ".sas": {"is_code": True, "language_mode": "text"},
    ".scss": {"is_code": False, "language_mode": "scss"},
    ".te": {"is_code": False, "language_mode": "text"},
    ".smt2": {"is_code": True, "language_mode": "text"},
    ".smt": {"is_code": True, "language_mode": "text"},
    ".sparql": {"is_code": False, "language_mode": "text"},
    ".rq": {"is_code": False, "language_mode": "text"},
    ".sqf": {"is_code": True, "language_mode": "text"},
    ".hqf": {"is_code": True, "language_mode": "text"},
    ".cql": {"is_code": False, "language_mode": "sql"},
    ".mysql": {"is_code": False, "language_mode": "sql"},
    ".tab": {"is_code": False, "language_mode": "sql"},
    ".udf": {"is_code": False, "language_mode": "sql"},
    ".viw": {"is_code": False, "language_mode": "sql"},
    ".db2": {"is_code": True, "language_mode": "sql"},
    ".srt": {"is_code": False, "language_mode": "text"},
    ".star": {"is_code": True, "language_mode": "python"},
    ".stl": {"is_code": False, "language_mode": "text"},
    ".ston": {"is_code": False, "language_mode": "text"},
    ".svg": {"is_code": False, "language_mode": "xml"},
    ".sage": {"is_code": True, "language_mode": "python"},
    ".sagews": {"is_code": True, "language_mode": "python"},
    ".sls": {"is_code": True, "language_mode": "scheme"},
    ".sass": {"is_code": False, "language_mode": "sass"},
    ".scala": {"is_code": True, "language_mode": "scala"},
    ".kojo": {"is_code": True, "language_mode": "scala"},
    ".sbt": {"is_code": True, "language_mode": "scala"},
    ".sc": {"is_code": True, "language_mode": "text"},
    ".scaml": {"is_code": False, "language_mode": "text"},
    ".scenic": {"is_code": True, "language_mode": "text"},
    ".scm": {"is_code": True, "language_mode": "scheme"},
    ".sld": {"is_code": True, "language_mode": "scheme"},
    ".sps": {"is_code": True, "language_mode": "scheme"},
    ".ss": {"is_code": True, "language_mode": "scheme"},
    ".sci": {"is_code": True, "language_mode": "text"},
    ".sce": {"is_code": True, "language_mode": "text"},
    ".self": {"is_code": True, "language_mode": "text"},
    ".sh": {"is_code": True, "language_mode": "sh"},
    ".bash": {"is_code": True, "language_mode": "sh"},
    ".bats": {"is_code": True, "language_mode": "sh"},
    ".command": {"is_code": True, "language_mode": "sh"},
    ".ksh": {"is_code": True, "language_mode": "sh"},
    ".sh.in": {"is_code": True, "language_mode": "sh"},
    ".tmux": {"is_code": True, "language_mode": "sh"},
    ".tool": {"is_code": True, "language_mode": "sh"},
    ".zsh": {"is_code": True, "language_mode": "sh"},
    ".zsh-theme": {"is_code": True, "language_mode": "sh"},
    ".sh-session": {"is_code": True, "language_mode": "sh"},
    ".shen": {"is_code": True, "language_mode": "text"},
    ".sieve": {"is_code": True, "language_mode": "text"},
    ".sfv": {"is_code": False, "language_mode": "ini"},
    ".sl": {"is_code": True, "language_mode": "text"},
    ".slim": {"is_code": False, "language_mode": "text"},
    ".cocci": {"is_code": True, "language_mode": "text"},
    ".smali": {"is_code": True, "language_mode": "text"},
    ".st": {"is_code": False, "language_mode": "html"},
    ".tpl": {"is_code": True, "language_mode": "smarty"},
    ".smithy": {"is_code": True, "language_mode": "text"},
    ".smk": {"is_code": True, "language_mode": "python"},
    ".snakefile": {"is_code": True, "language_mode": "python"},
    ".sp": {"is_code": True, "language_mode": "text"},
    ".sfd": {"is_code": False, "language_mode": "yaml"},
    ".nut": {"is_code": True, "language_mode": "c_cpp"},
    ".stan": {"is_code": True, "language_mode": "text"},
    ".fun": {"is_code": True, "language_mode": "text"},
    ".sig": {"is_code": True, "language_mode": "text"},
    ".sml": {"is_code": True, "language_mode": "text"},
    ".bzl": {"is_code": True, "language_mode": "python"},
    ".do": {"is_code": True, "language_mode": "text"},
    ".ado": {"is_code": True, "language_mode": "text"},
    ".doh": {"is_code": True, "language_mode": "text"},
    ".ihlp": {"is_code": True, "language_mode": "text"},
    ".mata": {"is_code": True, "language_mode": "text"},
    ".matah": {"is_code": True, "language_mode": "text"},
    ".sthlp": {"is_code": True, "language_mode": "text"},
    ".styl": {"is_code": False, "language_mode": "stylus"},
    ".sss": {"is_code": False, "language_mode": "text"},
    ".svelte": {"is_code": False, "language_mode": "html"},
    ".sw": {"is_code": False, "language_mode": "xml"},
    ".swift": {"is_code": True, "language_mode": "text"},
    ".sv": {"is_code": True, "language_mode": "verilog"},
    ".svh": {"is_code": True, "language_mode": "verilog"},
    ".vh": {"is_code": True, "language_mode": "verilog"},
    ".8xp": {"is_code": True, "language_mode": "text"},
    ".8xk": {"is_code": True, "language_mode": "text"},
    ".8xk.txt": {"is_code": True, "language_mode": "text"},
    ".8xp.txt": {"is_code": True, "language_mode": "text"},
    ".tla": {"is_code": True, "language_mode": "text"},
    ".toml": {"is_code": True, "language_mode": "toml"},
    ".tsv": {"is_code": False, "language_mode": "text"},
    ".tsx": {"is_code": True, "language_mode": "javascript"},
    ".txl": {"is_code": True, "language_mode": "text"},
    ".talon": {"is_code": True, "language_mode": "text"},
    ".tcl": {"is_code": True, "language_mode": "tcl"},
    ".adp": {"is_code": True, "language_mode": "tcl"},
    ".sdc": {"is_code": True, "language_mode": "tcl"},
    ".tcl.in": {"is_code": True, "language_mode": "tcl"},
    ".tm": {"is_code": True, "language_mode": "tcl"},
    ".xdc": {"is_code": True, "language_mode": "tcl"},
    ".tcsh": {"is_code": True, "language_mode": "sh"},
    ".csh": {"is_code": True, "language_mode": "sh"},
    ".tex": {"is_code": False, "language_mode": "tex"},
    ".aux": {"is_code": False, "language_mode": "tex"},
    ".bbx": {"is_code": False, "language_mode": "tex"},
    ".cbx": {"is_code": False, "language_mode": "tex"},
    ".dtx": {"is_code": False, "language_mode": "tex"},
    ".ins": {"is_code": False, "language_mode": "tex"},
    ".lbx": {"is_code": False, "language_mode": "tex"},
    ".ltx": {"is_code": False, "language_mode": "tex"},
    ".mkii": {"is_code": False, "language_mode": "tex"},
    ".mkiv": {"is_code": False, "language_mode": "tex"},
    ".mkvi": {"is_code": False, "language_mode": "tex"},
    ".sty": {"is_code": False, "language_mode": "tex"},
    ".toc": {"is_code": False, "language_mode": "text"},
    ".tea": {"is_code": False, "language_mode": "text"},
    ".texinfo": {"is_code": False, "language_mode": "text"},
    ".texi": {"is_code": False, "language_mode": "text"},
    ".txi": {"is_code": False, "language_mode": "text"},
    ".no": {"is_code": False, "language_mode": "text"},
    ".textile": {"is_code": False, "language_mode": "textile"},
    ".thrift": {"is_code": True, "language_mode": "text"},
    ".tu": {"is_code": True, "language_mode": "text"},
    ".ttl": {"is_code": False, "language_mode": "text"},
    ".twig": {"is_code": False, "language_mode": "twig"},
    ".tl": {"is_code": False, "language_mode": "text"},
    ".ts": {"is_code": True, "language_mode": "javascript"},
    ".cts": {"is_code": True, "language_mode": "typescript"},
    ".mts": {"is_code": True, "language_mode": "typescript"},
    ".upc": {"is_code": True, "language_mode": "c_cpp"},
    ".anim": {"is_code": False, "language_mode": "yaml"},
    ".asset": {"is_code": False, "language_mode": "yaml"},
    ".mat": {"is_code": False, "language_mode": "yaml"},
    ".meta": {"is_code": False, "language_mode": "yaml"},
    ".prefab": {"is_code": False, "language_mode": "yaml"},
    ".unity": {"is_code": False, "language_mode": "yaml"},
    ".uno": {"is_code": True, "language_mode": "csharp"},
    ".uc": {"is_code": True, "language_mode": "java"},
    ".ur": {"is_code": True, "language_mode": "text"},
    ".urs": {"is_code": True, "language_mode": "text"},
    ".frm": {"is_code": True, "language_mode": "text"},
    ".vba": {"is_code": True, "language_mode": "text"},
    ".vbs": {"is_code": True, "language_mode": "text"},
    ".vcl": {"is_code": True, "language_mode": "text"},
    ".vhdl": {"is_code": True, "language_mode": "vhdl"},
    ".vhd": {"is_code": True, "language_mode": "vhdl"},
    ".vhf": {"is_code": True, "language_mode": "vhdl"},
    ".vhi": {"is_code": True, "language_mode": "vhdl"},
    ".vho": {"is_code": True, "language_mode": "vhdl"},
    ".vhs": {"is_code": True, "language_mode": "vhdl"},
    ".vht": {"is_code": True, "language_mode": "vhdl"},
    ".vhw": {"is_code": True, "language_mode": "vhdl"},
    ".vala": {"is_code": True, "language_mode": "vala"},
    ".vapi": {"is_code": True, "language_mode": "vala"},
    ".vdf": {"is_code": False, "language_mode": "text"},
    ".vtl": {"is_code": False, "language_mode": "velocity"},
    ".veo": {"is_code": True, "language_mode": "verilog"},
    ".vim": {"is_code": True, "language_mode": "text"},
    ".vimrc": {"is_code": True, "language_mode": "text"},
    ".vmb": {"is_code": True, "language_mode": "text"},
    ".snip": {"is_code": False, "language_mode": "text"},
    ".snippet": {"is_code": False, "language_mode": "text"},
    ".snippets": {"is_code": False, "language_mode": "text"},
    ".vb": {"is_code": True, "language_mode": "text"},
    ".vbhtml": {"is_code": True, "language_mode": "text"},
    ".ctl": {"is_code": True, "language_mode": "text"},
    ".Dsr": {"is_code": True, "language_mode": "text"},
    ".volt": {"is_code": True, "language_mode": "d"},
    ".vue": {"is_code": False, "language_mode": "html"},
    ".vy": {"is_code": True, "language_mode": "text"},
    ".mtl": {"is_code": False, "language_mode": "text"},
    ".obj": {"is_code": False, "language_mode": "text"},
    ".owl": {"is_code": False, "language_mode": "xml"},
    ".wast": {"is_code": True, "language_mode": "lisp"},
    ".wat": {"is_code": True, "language_mode": "lisp"},
    ".webidl": {"is_code": True, "language_mode": "text"},
    ".vtt": {"is_code": False, "language_mode": "text"},
    ".whiley": {"is_code": True, "language_mode": "text"},
    ".mediawiki": {"is_code": False, "language_mode": "text"},
    ".wiki": {"is_code": False, "language_mode": "text"},
    ".wikitext": {"is_code": False, "language_mode": "text"},
    ".reg": {"is_code": False, "language_mode": "ini"},
    ".ws": {"is_code": True, "language_mode": "text"},
    ".wlk": {"is_code": True, "language_mode": "text"},
    ".wren": {"is_code": True, "language_mode": "text"},
    ".xbm": {"is_code": False, "language_mode": "c_cpp"},
    ".xpm": {"is_code": False, "language_mode": "c_cpp"},
    ".x10": {"is_code": True, "language_mode": "text"},
    ".xc": {"is_code": True, "language_mode": "c_cpp"},
    ".xml": {"is_code": False, "language_mode": "xml"},
    ".adml": {"is_code": False, "language_mode": "xml"},
    ".admx": {"is_code": False, "language_mode": "xml"},
    ".ant": {"is_code": False, "language_mode": "xml"},
    ".axaml": {"is_code": False, "language_mode": "xml"},
    ".axml": {"is_code": False, "language_mode": "xml"},
    ".builds": {"is_code": False, "language_mode": "xml"},
    ".ccproj": {"is_code": False, "language_mode": "xml"},
    ".ccxml": {"is_code": False, "language_mode": "xml"},
    ".clixml": {"is_code": False, "language_mode": "xml"},
    ".cproject": {"is_code": False, "language_mode": "xml"},
    ".cscfg": {"is_code": False, "language_mode": "xml"},
    ".csdef": {"is_code": False, "language_mode": "xml"},
    ".csproj": {"is_code": False, "language_mode": "xml"},
    ".ct": {"is_code": False, "language_mode": "xml"},
    ".depproj": {"is_code": False, "language_mode": "xml"},
    ".dita": {"is_code": False, "language_mode": "xml"},
    ".ditamap": {"is_code": False, "language_mode": "xml"},
    ".ditaval": {"is_code": False, "language_mode": "xml"},
    ".dll.config": {"is_code": False, "language_mode": "xml"},
    ".dotsettings": {"is_code": False, "language_mode": "xml"},
    ".filters": {"is_code": False, "language_mode": "xml"},
    ".fsproj": {"is_code": False, "language_mode": "xml"},
    ".fxml": {"is_code": False, "language_mode": "xml"},
    ".glade": {"is_code": False, "language_mode": "xml"},
    ".gmx": {"is_code": False, "language_mode": "xml"},
    ".grxml": {"is_code": False, "language_mode": "xml"},
    ".hzp": {"is_code": False, "language_mode": "xml"},
    ".iml": {"is_code": False, "language_mode": "xml"},
    ".ivy": {"is_code": False, "language_mode": "xml"},
    ".jelly": {"is_code": False, "language_mode": "xml"},
    ".jsproj": {"is_code": False, "language_mode": "xml"},
    ".kml": {"is_code": False, "language_mode": "xml"},
    ".launch": {"is_code": False, "language_mode": "xml"},
    ".mdpolicy": {"is_code": False, "language_mode": "xml"},
    ".mjml": {"is_code": False, "language_mode": "xml"},
    ".mxml": {"is_code": False, "language_mode": "xml"},
    ".natvis": {"is_code": False, "language_mode": "xml"},
    ".ndproj": {"is_code": False, "language_mode": "xml"},
    ".nproj": {"is_code": False, "language_mode": "xml"},
    ".nuspec": {"is_code": False, "language_mode": "xml"},
    ".odd": {"is_code": False, "language_mode": "xml"},
    ".osm": {"is_code": False, "language_mode": "xml"},
    ".pkgproj": {"is_code": False, "language_mode": "xml"},
    ".proj": {"is_code": False, "language_mode": "xml"},
    ".props": {"is_code": False, "language_mode": "xml"},
    ".ps1xml": {"is_code": False, "language_mode": "xml"},
    ".psc1": {"is_code": False, "language_mode": "xml"},
    ".pt": {"is_code": False, "language_mode": "xml"},
    ".qhelp": {"is_code": False, "language_mode": "xml"},
    ".rdf": {"is_code": False, "language_mode": "xml"},
    ".resx": {"is_code": False, "language_mode": "xml"},
    ".rss": {"is_code": False, "language_mode": "xml"},
    ".scxml": {"is_code": False, "language_mode": "xml"},
    ".sfproj": {"is_code": False, "language_mode": "xml"},
    ".shproj": {"is_code": False, "language_mode": "xml"},
    ".srdf": {"is_code": False, "language_mode": "xml"},
    ".storyboard": {"is_code": False, "language_mode": "xml"},
    ".sublime-snippet": {"is_code": False, "language_mode": "xml"},
    ".targets": {"is_code": False, "language_mode": "xml"},
    ".tml": {"is_code": True, "language_mode": "xml"},
    ".ui": {"is_code": False, "language_mode": "xml"},
    ".urdf": {"is_code": False, "language_mode": "xml"},
    ".ux": {"is_code": False, "language_mode": "xml"},
    ".vbproj": {"is_code": False, "language_mode": "xml"},
    ".vcxproj": {"is_code": False, "language_mode": "xml"},
    ".vsixmanifest": {"is_code": False, "language_mode": "xml"},
    ".vssettings": {"is_code": False, "language_mode": "xml"},
    ".vstemplate": {"is_code": False, "language_mode": "xml"},
    ".vxml": {"is_code": False, "language_mode": "xml"},
    ".wixproj": {"is_code": False, "language_mode": "xml"},
    ".wsdl": {"is_code": False, "language_mode": "xml"},
    ".wsf": {"is_code": False, "language_mode": "xml"},
    ".wxi": {"is_code": False, "language_mode": "xml"},
    ".wxl": {"is_code": False, "language_mode": "xml"},
    ".wxs": {"is_code": False, "language_mode": "xml"},
    ".x3d": {"is_code": False, "language_mode": "xml"},
    ".xacro": {"is_code": False, "language_mode": "xml"},
    ".xaml": {"is_code": False, "language_mode": "xml"},
    ".xib": {"is_code": False, "language_mode": "xml"},
    ".xlf": {"is_code": False, "language_mode": "xml"},
    ".xliff": {"is_code": False, "language_mode": "xml"},
    ".xmi": {"is_code": False, "language_mode": "xml"},
    ".xml.dist": {"is_code": False, "language_mode": "xml"},
    ".xmp": {"is_code": False, "language_mode": "xml"},
    ".xproj": {"is_code": False, "language_mode": "xml"},
    ".xsd": {"is_code": False, "language_mode": "xml"},
    ".xspec": {"is_code": False, "language_mode": "xml"},
    ".xul": {"is_code": False, "language_mode": "xml"},
    ".zcml": {"is_code": False, "language_mode": "xml"},
    ".stTheme": {"is_code": False, "language_mode": "xml"},
    ".tmCommand": {"is_code": False, "language_mode": "xml"},
    ".tmLanguage": {"is_code": False, "language_mode": "xml"},
    ".tmPreferences": {"is_code": False, "language_mode": "xml"},
    ".tmSnippet": {"is_code": False, "language_mode": "xml"},
    ".tmTheme": {"is_code": False, "language_mode": "xml"},
    ".xsp-config": {"is_code": False, "language_mode": "xml"},
    ".xsp.metadata": {"is_code": False, "language_mode": "xml"},
    ".xpl": {"is_code": True, "language_mode": "xml"},
    ".xproc": {"is_code": True, "language_mode": "xml"},
    ".xquery": {"is_code": True, "language_mode": "xquery"},
    ".xq": {"is_code": True, "language_mode": "xquery"},
    ".xql": {"is_code": True, "language_mode": "xquery"},
    ".xqm": {"is_code": True, "language_mode": "xquery"},
    ".xqy": {"is_code": True, "language_mode": "xquery"},
    ".xs": {"is_code": True, "language_mode": "c_cpp"},
    ".xslt": {"is_code": True, "language_mode": "xml"},
    ".xsl": {"is_code": True, "language_mode": "xml"},
    ".xojo_code": {"is_code": True, "language_mode": "text"},
    ".xojo_menu": {"is_code": True, "language_mode": "text"},
    ".xojo_report": {"is_code": True, "language_mode": "text"},
    ".xojo_script": {"is_code": True, "language_mode": "text"},
    ".xojo_toolbar": {"is_code": True, "language_mode": "text"},
    ".xojo_window": {"is_code": True, "language_mode": "text"},
    ".xsh": {"is_code": True, "language_mode": "text"},
    ".xtend": {"is_code": True, "language_mode": "text"},
    ".mir": {"is_code": False, "language_mode": "yaml"},
    ".reek": {"is_code": False, "language_mode": "yaml"},
    ".rviz": {"is_code": False, "language_mode": "yaml"},
    ".sublime-syntax": {"is_code": False, "language_mode": "yaml"},
    ".syntax": {"is_code": False, "language_mode": "yaml"},
    ".yaml-tmlanguage": {"is_code": False, "language_mode": "yaml"},
    ".yaml.sed": {"is_code": False, "language_mode": "yaml"},
    ".yml.mysql": {"is_code": False, "language_mode": "yaml"},
    ".yang": {"is_code": False, "language_mode": "text"},
    ".yar": {"is_code": True, "language_mode": "text"},
    ".yara": {"is_code": True, "language_mode": "text"},
    ".yasnippet": {"is_code": False, "language_mode": "text"},
    ".y": {"is_code": True, "language_mode": "text"},
    ".yacc": {"is_code": True, "language_mode": "text"},
    ".yul": {"is_code": True, "language_mode": "text"},
    ".zap": {"is_code": True, "language_mode": "text"},
    ".xzap": {"is_code": True, "language_mode": "text"},
    ".zil": {"is_code": True, "language_mode": "text"},
    ".mud": {"is_code": True, "language_mode": "text"},
    ".zeek": {"is_code": True, "language_mode": "text"},
    ".bro": {"is_code": True, "language_mode": "text"},
    ".zs": {"is_code": True, "language_mode": "text"},
    ".zep": {"is_code": True, "language_mode": "php"},
    ".zig": {"is_code": True, "language_mode": "text"},
    ".zimpl": {"is_code": True, "language_mode": "text"},
    ".zmpl": {"is_code": True, "language_mode": "text"},
    ".zpl": {"is_code": True, "language_mode": "text"},
    ".desktop": {"is_code": False, "language_mode": "text"},
    ".desktop.in": {"is_code": False, "language_mode": "text"},
    ".service": {"is_code": False, "language_mode": "text"},
    ".dircolors": {"is_code": False, "language_mode": "text"},
    ".ec": {"is_code": True, "language_mode": "text"},
    ".eh": {"is_code": True, "language_mode": "text"},
    ".edn": {"is_code": False, "language_mode": "clojure"},
    ".fish": {"is_code": True, "language_mode": "text"},
    ".hoon": {"is_code": True, "language_mode": "text"},
    ".kv": {"is_code": False, "language_mode": "text"},
    ".mrc": {"is_code": True, "language_mode": "text"},
    ".mcfunction": {"is_code": True, "language_mode": "text"},
    ".mu": {"is_code": True, "language_mode": "text"},
    ".nanorc": {"is_code": False, "language_mode": "text"},
    ".nc": {"is_code": True, "language_mode": "text"},
    ".ooc": {"is_code": True, "language_mode": "text"},
    ".rst": {"is_code": False, "language_mode": "text"},
    ".rest": {"is_code": False, "language_mode": "text"},
    ".rest.txt": {"is_code": False, "language_mode": "text"},
    ".rst.txt": {"is_code": False, "language_mode": "text"},
    ".sed": {"is_code": True, "language_mode": "text"},
    ".wdl": {"is_code": True, "language_mode": "text"},
    ".wisp": {"is_code": True, "language_mode": "clojure"},
    ".prg": {"is_code": True, "language_mode": "text"},
    ".prw": {"is_code": True, "language_mode": "text"},
}



from withrepo.withrepo import (
    repo,
    RepoContext,
    RepoFile,
)

from withrepo.utils import (
    RepoProvider,
    RepoArguments,
)

from withrepo.download import (
    copy_and_split_root_by_language_group,
)

__all__ = [
    "repo",
    "RepoContext",
    "RepoFile",
    "File",
    "RepoArguments",
    "RepoProvider",
    "copy_and_split_root_by_language_group",
]


# TODO: Artifact from scope, remove eventually when API is stable
LANGUAGE_TO_LSP_LANGUAGE_MAP = {
    "python": "python",
    "javascript": "typescript",
    "typescript": "typescript",
    "java": "java",
    "rust": "rust",
    "csharp": "csharp",
    "golang": "go",
    "ruby": "ruby",
}



"""
Attribution:
    download_and_extract_archive():
        https://github.com/microsoft/multilspy/blob/main/src/multilspy/multilspy_utils.py

    Original authors:
        Lakshya A Agrawal

    License:
        MIT
"""

# Standard library
import os
import shutil
import tempfile
from typing import List, Tuple

# Local
from withrepo.utils import (
    RepoArguments,
    RepoProvider,
    LanguageGroup,
    copy_and_split_root_by_language_group,
)

# Third party
import httpx

# CONSTANTS
CHUNK_SIZE = 2 * 1024 * 1024

PROVIDER_TO_URL_MAP = {
    RepoProvider.GITHUB: "https://github.com",
    RepoProvider.GITLAB: "https://gitlab.com",
    RepoProvider.BITBUCKET: "https://bitbucket.org",
}


def download_and_extract_archive(url: str) -> Tuple[str, List[LanguageGroup]]:
    """
    Downloads the archive from the given URL having format {archive_type} and extracts it to the given {target_path}
    """
    if not url:
        raise Exception("withrepo.download_file(): URL is empty")

    archive_type = url.split(".")[-1]
    extract_directory = tempfile.mkdtemp(prefix="scope_")
    fd, tmp_file_name = tempfile.mkstemp(prefix="scope_")
    lang_groups = []

    try:
        # Download the archive
        client = httpx.Client(follow_redirects=True, http2=True)
        with client.stream("GET", url, timeout=60.0) as response:
            if response.status_code != 200:
                error_text = response.read().decode()
                raise httpx.RequestError(
                    f"Error downloading file '{url}': {response.status_code} {error_text}"
                )
            with os.fdopen(fd, "wb") as f:
                for chunk in response.iter_raw(chunk_size=CHUNK_SIZE):
                    f.write(chunk)

        # Extract the archive
        if archive_type in {"zip", "tar", "gztar", "bztar", "xztar"}:
            shutil.unpack_archive(tmp_file_name, extract_directory, archive_type)
        else:
            raise Exception(
                f"download_and_extract_archive(): Unsupported archive type '{archive_type}'"
            )

        # Split the archive into language groups
        lang_groups.extend(copy_and_split_root_by_language_group(extract_directory))

        # TODO: figure out if this screws up with path removal on cleanup
        child_dirs = os.listdir(extract_directory)
        if len(child_dirs) == 1:
            extract_directory = os.path.join(extract_directory, child_dirs[0])
    except Exception as exc:
        raise Exception(
            f"Error extracting archive '{tmp_file_name}' obtained from '{url}': {exc}"
        ) from exc
    finally:
        if os.path.exists(tmp_file_name):
            os.remove(tmp_file_name)
        # if os.path.exists(extract_directory):
        #     shutil.rmtree(extract_directory)

    return extract_directory, lang_groups


### TODO
# GITLAB default clone
# https://gitlab.com/NTPsec/ntpsec/-/archive/master/ntpsec-master.zip

# GITLAB commit clone
# https://gitlab.com/NTPsec/ntpsec/-/archive/f282438ee3875330d82877975adbf7a096b06f73/ntpsec-f282438ee3875330d82877975adbf7a096b06f73.zip

# GITLAB branch clone
# https://gitlab.com/NTPsec/ntpsec/-/archive/mr_1415/ntpsec-mr_1415.zip


def parse_repo_arguments_into_download_url(args: RepoArguments) -> str:
    if args.invalid():
        raise Exception("Cannot parse repo() without arguments")
    provider_url = PROVIDER_TO_URL_MAP[args.provider]
    # at minimum, either url or (user, repo) must be provided
    if args.url:
        if not args.commit:
            # if args.provider == RepoProvider.GITLAB:
            #     return f"{provider_url}/{args.user}/{args.repo}/-/raw/HEAD/archive.zip"
            return f"{args.url}/archive/HEAD.zip"
        else:
            return f"{args.url}/archive/{args.commit}.zip"
    elif args.user and args.repo:
        if args.branch:
            return f"{provider_url}/{args.user}/{args.repo}/archive/{args.branch}.zip"
        elif args.commit:
            return f"{provider_url}/{args.user}/{args.repo}/archive/{args.commit}.zip"
        else:
            return f"{provider_url}/{args.user}/{args.repo}/archive/HEAD.zip"
    else:
        raise Exception("Cannot parse repo() with given arguments")



from withrepo.resources.languages import EXT_TO_LANGUAGE_DATA
from withrepo.resources.tree_sitter import EXT_TO_TREE_SITTER_LANGUAGE

__all__ = ["EXT_TO_LANGUAGE_DATA", "EXT_TO_TREE_SITTER_LANGUAGE"]



# Standard library
import os
import shutil
import contextlib
from typing import Iterator, List, Union, Dict
from collections import defaultdict

# Local
from withrepo.utils import RepoArguments, LanguageGroup, RepoProvider
from withrepo.download import (
    parse_repo_arguments_into_download_url,
    download_and_extract_archive,
)
from withrepo.utils import (
    get_language_from_ext,
    copy_and_split_root_by_language_group
) 

from withrepo.resources.languages import EXT_TO_LANGUAGE_DATA


# THIS IS JUST HERE FOR COMPATIBILITY WITH ADRENALINE PROD
class File(object):
    def __init__(
        self,
        path,
        abs_path,
        content=None,
        language=None,
        summary=None,
        has_ingress=False,
        change_type=None,
        db_id=None,
        **kwargs,
    ):
        self.db_id = db_id
        self.path = path  # Relative to the root directory
        self.abs_path = abs_path  # Relative to the repository root
        self.content = content
        self.change_type = change_type
        self.summary = summary

        self.file_name = File.get_file_name(path)
        self.file_extension = File.get_file_extension(path)

        if language:
            self.language = language
        elif self.file_extension in EXT_TO_LANGUAGE_DATA:
            self.language = EXT_TO_LANGUAGE_DATA[self.file_extension]["language_mode"]
        else:
            self.language = None

        self.embedding = None
        self.children = []
        self.has_ingress = has_ingress

    def __repr__(self):
        return f"File(path={self.path})"

    @staticmethod
    def get_file_name(path):
        if not path:
            return None

        path_components = path.split("/")
        file_name = path_components[-1]
        return file_name

    @staticmethod
    def get_file_extension(path):
        if not path:
            return None

        file_name = File.get_file_name(path)
        if file_name.lower() == "readme":
            return ".md"

        return f".{file_name.split('.')[-1]}".lower()

    def to_dict(self):
        return {
            "path": self.path,
            "abs_path": self.abs_path,
            "content": self.content,
            "language": self.language,
            "summary": self.summary,
            "has_ingress": self.has_ingress,
        }

    def to_string(self):
        return f"<FILE_PATH>{self.path}</FILE_PATH>\n<FILE_SUMMARY>\n{self.summary}\n</FILE_SUMMARY>"


class RepoFile:
    def __init__(self, abs_path: str, relative_path: str, preload: bool = False):
        self.file_name: str = os.path.basename(abs_path)
        self.file_extension: str = os.path.splitext(abs_path)[1]
        self.abs_path: str = abs_path
        self.path: str = relative_path
        self._contents: str = None
        if preload:
            self.contents()
        _, language, is_code = get_language_from_ext(abs_path)
        self.language: str = language
        self.is_code: bool = is_code
        # self.tree_sitter_lang = EXT_TO_TREE_SITTER_LANGUAGE.get(self.ext, None)

    def __len__(self) -> int:
        return len(self.content.split("\n"))

    def __str__(self) -> str:
        return f"RepoFile(file_name={self.file_name}, path={self.path}, language={self.language}, is_code={self.is_code}, num_lines={len(self)})"

    def __iter__(self) -> Iterator[str]:
        return iter(self.content.split("\n"))

    @property
    def content(self) -> str:
        return self.contents()

    def contents(self) -> str:
        try:
            if self._contents is None:
                with open(self.abs_path, "r") as f:
                    self._contents = f.read()
            return self._contents
        except Exception:
            # print(f"RepoFile::contents() Error reading file {self.abs_path}")
            return ""


class RepoContext:
    def __init__(
        self, path: str, url: str, args: RepoArguments, lang_groups: List[LanguageGroup]
    ):
        """Stores the context for a withrepo test."""
        self.path: str = path
        self.url: str = url
        self.user: str = args.user
        self.repo: str = args.repo
        self.commit: str = args.commit
        self.branch: str = args.branch
        self.repo_url: str = args.url
        self.root_dir: str = args.root_dir
        self.provider: RepoProvider = args.provider
        self.lang_groups: List[LanguageGroup] = lang_groups
        self.languages: List[str] = list(
            {lang_group.language for lang_group in lang_groups}
        )

        self.files: List[RepoFile] = []
        self.lang_trees: Dict[str, List[RepoFile]] = {}

    def __str__(self):
        return f"""RepoContext(
            url={self.url},
            user={self.user},
            repo={self.repo},
            commit={self.commit},
            branch={self.branch},
            provider={self.provider},
            languages={self.languages},
            path={self.path}
        )"""

    def tree(
        self, multilang: bool = False, store: bool = False
    ) -> Union[List[RepoFile], Dict[str, List[RepoFile]]]:
        """
        Returns a tree of the repository.
        If multilang is False, returns a list of RepoFiles.
        If multilang is True, returns a dict mapping languages to lists of RepoFiles.
        """
        if not multilang:
            files = []
            for root, _, files_list in os.walk(self.path, topdown=True):
                # Collect full paths relative to the input directory
                for file in files_list:
                    relpath = os.path.relpath(os.path.join(root, file), self.path)
                    abspath = os.path.abspath(os.path.join(root, file))
                    if self.root_dir and not relpath.startswith(self.root_dir):
                        continue
                    files.append(RepoFile(abspath, relpath, preload=store))
            if store:
                self.files = files
            return files
        else:
            lang_trees = defaultdict(list)
            for lang_group in self.lang_groups:
                lang_tree = list(os.walk(lang_group.path, topdown=True))
                tree_files_and_root = [(root, files) for root, _, files in lang_tree]
                for root, files in tree_files_and_root:
                    for file in files:
                        relpath = os.path.relpath(
                            os.path.join(root, file), lang_group.path
                        )
                        abspath = os.path.abspath(os.path.join(root, file))
                        if self.root_dir and not relpath.startswith(self.root_dir):
                            continue
                        lang_trees[lang_group.language].append(
                            RepoFile(abspath, relpath, preload=store)
                        )
            if store:
                self.lang_trees = lang_trees
            return lang_trees

    def cleanup(self, log: bool = False):
        if log:
            print(f"RepoContext::cleanup() Cleaning up {self.path}")
            print(f"RepoContext::cleanup() Cleaning up {self.lang_groups}")
        # cleanup the source directory and the group directories
        if os.path.exists(self.path):
            shutil.rmtree(self.path)
        for lang_group in self.lang_groups:
            if os.path.exists(lang_group.path):
                shutil.rmtree(lang_group.path)


@contextlib.contextmanager
def repo(
    user: str = None,
    repo: str = None,
    commit: str = "",
    branch: str = "",
    url: str = "",
    root_dir: str = "",
    root_path: str = None,
    provider: RepoProvider = RepoProvider.GITHUB,
    cleanup_callback: bool = False,
    timeit: bool = False,
    log: bool = False,
) -> Iterator[RepoContext]:
    args = RepoArguments(
        user=user,
        repo=repo,
        commit=commit,
        branch=branch,
        url=url,
        provider=provider,
        root_dir=root_dir,
        root_path=root_path,
    )

    if args.invalid():
        raise ValueError("Invalid repo arguments")

    if not root_path:
        repo_zip_url = parse_repo_arguments_into_download_url(args)
        source_directory_path, lang_groups = download_and_extract_archive(repo_zip_url)
    else:
        repo_zip_url = None
        lang_groups = copy_and_split_root_by_language_group(root_path)
        source_directory_path = root_path

    repo_ctx = RepoContext(source_directory_path, repo_zip_url, args, lang_groups)
    yield repo_ctx

    if not root_path and not cleanup_callback:
        repo_ctx.cleanup(log=log)
